# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Rational-quadratic spline bijector."""

from typing import Tuple

from distrax._src.bijectors import bijector as base
import jax
import jax.numpy as jnp


Array = base.Array


def _normalize_bin_sizes(unnormalized_bin_sizes: Array,
                         total_size: float,
                         min_bin_size: float) -> Array:
  """Make bin sizes sum to `total_size` and be no less than `min_bin_size`."""
  num_bins = unnormalized_bin_sizes.shape[-1]
  if num_bins * min_bin_size > total_size:
    raise ValueError(
        f'The number of bins ({num_bins}) times the minimum bin size'
        f' ({min_bin_size}) cannot be greater than the total bin size'
        f' ({total_size}).')
  bin_sizes = jax.nn.softmax(unnormalized_bin_sizes, axis=-1)
  return bin_sizes * (total_size - num_bins * min_bin_size) + min_bin_size


def _normalize_knot_slopes(unnormalized_knot_slopes: Array,
                           min_knot_slope: float) -> Array:
  """Make knot slopes be no less than `min_knot_slope`."""
  # The offset is such that the normalized knot slope will be equal to 1
  # whenever the unnormalized knot slope is equal to 0.
  if min_knot_slope >= 1.:
    raise ValueError(f'The minimum knot slope must be less than 1; got'
                     f' {min_knot_slope}.')
  min_knot_slope = jnp.array(
      min_knot_slope, dtype=unnormalized_knot_slopes.dtype)
  offset = jnp.log(jnp.exp(1. - min_knot_slope) - 1.)
  return jax.nn.softplus(unnormalized_knot_slopes + offset) + min_knot_slope


def _rational_quadratic_spline_fwd(x: Array,
                                   x_pos: Array,
                                   y_pos: Array,
                                   knot_slopes: Array) -> Tuple[Array, Array]:
  """Applies a rational-quadratic spline to a scalar.

  Args:
    x: a scalar (0-dimensional array). The scalar `x` can be any real number; it
      will be transformed by the spline if it's in the closed interval
      `[x_pos[0], x_pos[-1]]`, and it will be transformed linearly if it's
      outside that interval.
    x_pos: array of shape [num_bins + 1], the bin boundaries on the x axis.
    y_pos: array of shape [num_bins + 1], the bin boundaries on the y axis.
    knot_slopes: array of shape [num_bins + 1], the slopes at the knot points.
  Returns:
    A tuple of two scalars: the output of the transformation and the log of the
    absolute first derivative at `x`.
  """
  # Search to find the right bin. NOTE: The bins are sorted, so we could use
  # binary search, but this is more GPU/TPU friendly.
  # The following implementation avoids indexing for faster TPU computation.
  below_range = x <= x_pos[0]
  above_range = x >= x_pos[-1]
  correct_bin = jnp.logical_and(x >= x_pos[:-1], x < x_pos[1:])
  any_bin_in_range = jnp.any(correct_bin)
  first_bin = jnp.concatenate([jnp.array([1]),
                               jnp.zeros(len(correct_bin)-1)]).astype(bool)
  # If y does not fall into any bin, we use the first spline in the following
  # computations to avoid numerical issues.
  correct_bin = jnp.where(any_bin_in_range, correct_bin, first_bin)
  # Dot product of each parameter with the correct bin mask.
  params = jnp.stack([x_pos, y_pos, knot_slopes], axis=1)
  params_bin_left = jnp.sum(correct_bin[:, None] * params[:-1], axis=0)
  params_bin_right = jnp.sum(correct_bin[:, None] * params[1:], axis=0)

  x_pos_bin = (params_bin_left[0], params_bin_right[0])
  y_pos_bin = (params_bin_left[1], params_bin_right[1])
  knot_slopes_bin = (params_bin_left[2], params_bin_right[2])

  bin_width = x_pos_bin[1] - x_pos_bin[0]
  bin_height = y_pos_bin[1] - y_pos_bin[0]
  bin_slope = bin_height / bin_width

  z = (x - x_pos_bin[0]) / bin_width
  # `z` should be in range [0, 1] to avoid NaNs later. This can happen because
  # of small floating point issues or when x is outside of the range of bins.
  # To avoid all problems, we restrict z in [0, 1].
  z = jnp.clip(z, 0., 1.)
  sq_z = z * z
  z1mz = z - sq_z  # z(1-z)
  sq_1mz = (1. - z) ** 2
  slopes_term = knot_slopes_bin[1] + knot_slopes_bin[0] - 2. * bin_slope
  numerator = bin_height * (bin_slope * sq_z + knot_slopes_bin[0] * z1mz)
  denominator = bin_slope + slopes_term * z1mz
  y = y_pos_bin[0] + numerator / denominator

  # Compute log det Jacobian.
  # The logdet is a sum of 3 logs. It is easy to see that the inputs of the
  # first two logs are guaranteed to be positive because we ensured that z is in
  # [0, 1]. This is also true of the log(denominator) because:
  # denominator
  # == bin_slope + (knot_slopes_bin[1] + knot_slopes_bin[0] - 2 * bin_slope) *
  # z*(1-z)
  # >= bin_slope - 2 * bin_slope * z * (1-z)
  # >= bin_slope - 2 * bin_slope * (1/4)
  # == bin_slope / 2
  logdet = 2. * jnp.log(bin_slope) + jnp.log(
      knot_slopes_bin[1] * sq_z + 2. * bin_slope * z1mz +
      knot_slopes_bin[0] * sq_1mz) - 2. * jnp.log(denominator)

  # If x is outside the spline range, we default to a linear transformation.
  y = jnp.where(below_range, (x - x_pos[0]) * knot_slopes[0] + y_pos[0], y)
  y = jnp.where(above_range, (x - x_pos[-1]) * knot_slopes[-1] + y_pos[-1], y)
  logdet = jnp.where(below_range, jnp.log(knot_slopes[0]), logdet)
  logdet = jnp.where(above_range, jnp.log(knot_slopes[-1]), logdet)
  return y, logdet


def _safe_quadratic_root(a: Array, b: Array, c: Array) -> Array:
  """Implement a numerically stable version of the quadratic formula."""
  # This is not a general solution to the quadratic equation, as it assumes
  # b ** 2 - 4. * a * c is known a priori to be positive (and which of the two
  # roots is to be used, see https://arxiv.org/abs/1906.04032).
  # There are two sources of instability:
  # (a) When b ** 2 - 4. * a * c -> 0, sqrt gives NaNs in gradient.
  # We clip sqrt_diff to have the smallest float number.
  sqrt_diff = b ** 2 - 4. * a * c
  safe_sqrt = jnp.sqrt(jnp.clip(sqrt_diff, jnp.finfo(sqrt_diff.dtype).tiny))
  # If sqrt_diff is non-positive, we set sqrt to 0. as it should be positive.
  safe_sqrt = jnp.where(sqrt_diff > 0., safe_sqrt, 0.)
  # (b) When 4. * a * c -> 0. We use the more stable quadratic solution
  # depending on the sign of b.
  # See https://people.csail.mit.edu/bkph/articles/Quadratics.pdf (eq 7 and 8).
  # Solution when b >= 0
  numerator_1 = 2. * c
  denominator_1 = -b - safe_sqrt
  # Solution when b < 0
  numerator_2 = - b + safe_sqrt
  denominator_2 = 2 * a
  # Choose the numerically stable solution.
  numerator = jnp.where(b >= 0, numerator_1, numerator_2)
  denominator = jnp.where(b >= 0, denominator_1, denominator_2)
  return numerator / denominator


def _rational_quadratic_spline_inv(y: Array,
                                   x_pos: Array,
                                   y_pos: Array,
                                   knot_slopes: Array) -> Tuple[Array, Array]:
  """Applies the inverse of a rational-quadratic spline to a scalar.

  Args:
    y: a scalar (0-dimensional array). The scalar `y` can be any real number; it
      will be transformed by the spline if it's in the closed interval
      `[y_pos[0], y_pos[-1]]`, and it will be transformed linearly if it's
      outside that interval.
    x_pos: array of shape [num_bins + 1], the bin boundaries on the x axis.
    y_pos: array of shape [num_bins + 1], the bin boundaries on the y axis.
    knot_slopes: array of shape [num_bins + 1], the slopes at the knot points.
  Returns:
    A tuple of two scalars: the output of the inverse transformation and the log
    of the absolute first derivative of the inverse at `y`.
  """
  # Search to find the right bin. NOTE: The bins are sorted, so we could use
  # binary search, but this is more GPU/TPU friendly.
  # The following implementation avoids indexing for faster TPU computation.
  below_range = y <= y_pos[0]
  above_range = y >= y_pos[-1]
  correct_bin = jnp.logical_and(y >= y_pos[:-1], y < y_pos[1:])
  any_bin_in_range = jnp.any(correct_bin)
  first_bin = jnp.concatenate([jnp.array([1]),
                               jnp.zeros(len(correct_bin)-1)]).astype(bool)
  # If y does not fall into any bin, we use the first spline in the following
  # computations to avoid numerical issues.
  correct_bin = jnp.where(any_bin_in_range, correct_bin, first_bin)
  # Dot product of each parameter with the correct bin mask.
  params = jnp.stack([x_pos, y_pos, knot_slopes], axis=1)
  params_bin_left = jnp.sum(correct_bin[:, None] * params[:-1], axis=0)
  params_bin_right = jnp.sum(correct_bin[:, None] * params[1:], axis=0)

  # These are the parameters for the corresponding bin.
  x_pos_bin = (params_bin_left[0], params_bin_right[0])
  y_pos_bin = (params_bin_left[1], params_bin_right[1])
  knot_slopes_bin = (params_bin_left[2], params_bin_right[2])

  bin_width = x_pos_bin[1] - x_pos_bin[0]
  bin_height = y_pos_bin[1] - y_pos_bin[0]
  bin_slope = bin_height / bin_width
  w = (y - y_pos_bin[0]) / bin_height
  w = jnp.clip(w, 0., 1.)  # Ensure w is in [0, 1].
  # Compute quadratic coefficients: az^2 + bz + c = 0
  slopes_term = knot_slopes_bin[1] + knot_slopes_bin[0] - 2. * bin_slope
  c = - bin_slope * w
  b = knot_slopes_bin[0] - slopes_term * w
  a = bin_slope - b

  # Solve quadratic to obtain z and then x.
  z = _safe_quadratic_root(a, b, c)
  z = jnp.clip(z, 0., 1.)  # Ensure z is in [0, 1].
  x = bin_width * z + x_pos_bin[0]

  # Compute log det Jacobian.
  sq_z = z * z
  z1mz = z - sq_z  # z(1-z)
  sq_1mz = (1. - z) ** 2
  denominator = bin_slope + slopes_term * z1mz
  logdet = - 2. * jnp.log(bin_slope) - jnp.log(
      knot_slopes_bin[1] * sq_z + 2. * bin_slope * z1mz +
      knot_slopes_bin[0] * sq_1mz) + 2. * jnp.log(denominator)

  # If y is outside the spline range, we default to a linear transformation.
  x = jnp.where(below_range, (y - y_pos[0]) / knot_slopes[0] + x_pos[0], x)
  x = jnp.where(above_range, (y - y_pos[-1]) / knot_slopes[-1] + x_pos[-1], x)
  logdet = jnp.where(below_range, - jnp.log(knot_slopes[0]), logdet)
  logdet = jnp.where(above_range, - jnp.log(knot_slopes[-1]), logdet)
  return x, logdet


class RationalQuadraticSpline(base.Bijector):
  """A rational-quadratic spline bijector.

  Implements the spline bijector introduced by:
  > Durkan et al., Neural Spline Flows, https://arxiv.org/abs/1906.04032, 2019.

  This bijector is a monotonically increasing spline operating on an interval
  [a, b], such that f(a) = a and f(b) = b. Outside the interval [a, b], the
  bijector defaults to a linear transformation whose slope matches that of the
  spline at the nearest boundary (either a or b). The range boundaries a and b
  are hyperparameters passed to the constructor.

  The spline on the interval [a, b] consists of `num_bins` segments, on each of
  which the spline takes the form of a rational quadratic (ratio of two
  quadratic polynomials). The first derivative of the bijector is guaranteed to
  be continuous on the whole real line. The second derivative is generally not
  continuous at the knot points (bin boundaries).

  The spline is parameterized by the bin sizes on the x and y axis, and by the
  slopes at the knot points. All spline parameters are passed to the constructor
  as an unconstrained array `params` of shape `[..., 3 * num_bins + 1]`. The
  spline parameters are extracted from `params`, and are reparameterized
  internally as appropriate. The number of bins is a hyperparameter, and is
  implicitly defined by the last dimension of `params`.

  This bijector is applied elementwise. Given some input `x`, the parameters
  `params` and the input `x` are broadcast against each other. For example,
  suppose `x` is of shape `[N, D]`. Then:
  - If `params` is of shape `[3 * num_bins + 1]`, the same spline is identically
    applied to each element of `x`.
  - If `params` is of shape `[D, 3 * num_bins + 1]`, the same spline is applied
    along the first axis of `x` but a different spline is applied along the
    second axis of `x`.
  - If `params` is of shape `[N, D, 3 * num_bins + 1]`, a different spline is
    applied to each element of `x`.
  - If `params` is of shape `[M, N, D, 3 * num_bins + 1]`, `M` different splines
    are applied to each element of `x`, and the output is of shape `[M, N, D]`.
  """

  def __init__(self,
               params: Array,
               range_min: float,
               range_max: float,
               boundary_slopes: str = 'unconstrained',
               min_bin_size: float = 1e-4,
               min_knot_slope: float = 1e-4):
    """Initializes a RationalQuadraticSpline bijector.

    Args:
      params: array of shape `[..., 3 * num_bins + 1]`, the unconstrained
        parameters of the bijector. The number of bins is implicitly defined by
        the last dimension of `params`. The parameters can take arbitrary
        unconstrained values; the bijector will reparameterize them internally
        and make sure they obey appropriate constraints. If `params` is the
        all-zeros array, the bijector becomes the identity function everywhere
        on the real line.
      range_min: the lower bound of the spline's range. Below `range_min`, the
        bijector defaults to a linear transformation.
      range_max: the upper bound of the spline's range. Above `range_max`, the
        bijector defaults to a linear transformation.
      boundary_slopes: controls the behaviour of the slope of the spline at the
        range boundaries (`range_min` and `range_max`). It is used to enforce
        certain boundary conditions on the spline. Available options are:
        - 'unconstrained': no boundary conditions are imposed; the slopes at the
          boundaries can vary freely.
        - 'lower_identity': the slope of the spline is set equal to 1 at the
          lower boundary (`range_min`). This makes the bijector equal to the
          identity function for values less than `range_min`.
        - 'upper_identity': similar to `lower_identity`, but now the slope of
          the spline is set equal to 1 at the upper boundary (`range_max`). This
          makes the bijector equal to the identity function for values greater
          than `range_max`.
        - 'identity': combines the effects of 'lower_identity' and
          'upper_identity' together. The slope of the spline is set equal to 1
          at both boundaries (`range_min` and `range_max`). This makes the
          bijector equal to the identity function outside the interval
          `[range_min, range_max]`.
        - 'circular': makes the slope at `range_min` and `range_max` be the
          same. This implements the "circular spline" introduced by:
          > Rezende et al., Normalizing Flows on Tori and Spheres,
          > https://arxiv.org/abs/2002.02428, 2020.
          This option should be used when the spline operates on a circle
          parameterized by an angle in the interval `[range_min, range_max]`,
          where `range_min` and `range_max` correspond to the same point on the
          circle.
      min_bin_size: The minimum bin size, in either the x or the y axis. Should
        be a small positive number, chosen for numerical stability. Guarantees
        that no bin in either the x or the y axis will be less than this value.
      min_knot_slope: The minimum slope at each knot point. Should be a small
        positive number, chosen for numerical stability. Guarantess that no knot
        will have a slope less than this value.
    """
    super().__init__(event_ndims_in=0)
    if params.shape[-1] % 3 != 1 or params.shape[-1] < 4:
      raise ValueError(f'The last dimension of `params` must have size'
                       f' `3 * num_bins + 1` and `num_bins` must be at least 1.'
                       f' Got size {params.shape[-1]}.')
    if range_min >= range_max:
      raise ValueError(f'`range_min` must be less than `range_max`. Got'
                       f' `range_min={range_min}` and `range_max={range_max}`.')
    if min_bin_size <= 0.:
      raise ValueError(f'The minimum bin size must be positive; got'
                       f' {min_bin_size}.')
    if min_knot_slope <= 0.:
      raise ValueError(f'The minimum knot slope must be positive; got'
                       f' {min_knot_slope}.')
    self._dtype = params.dtype
    self._num_bins = (params.shape[-1] - 1) // 3
    # Extract unnormalized parameters.
    unnormalized_bin_widths = params[..., :self._num_bins]
    unnormalized_bin_heights = params[..., self._num_bins : 2 * self._num_bins]
    unnormalized_knot_slopes = params[..., 2 * self._num_bins:]
    # Normalize bin sizes and compute bin positions on the x and y axis.
    range_size = range_max - range_min
    bin_widths = _normalize_bin_sizes(unnormalized_bin_widths, range_size,
                                      min_bin_size)
    bin_heights = _normalize_bin_sizes(unnormalized_bin_heights, range_size,
                                       min_bin_size)
    x_pos = range_min + jnp.cumsum(bin_widths[..., :-1], axis=-1)
    y_pos = range_min + jnp.cumsum(bin_heights[..., :-1], axis=-1)
    pad_shape = params.shape[:-1] + (1,)
    pad_below = jnp.full(pad_shape, range_min, dtype=self._dtype)
    pad_above = jnp.full(pad_shape, range_max, dtype=self._dtype)
    self._x_pos = jnp.concatenate([pad_below, x_pos, pad_above], axis=-1)
    self._y_pos = jnp.concatenate([pad_below, y_pos, pad_above], axis=-1)
    # Normalize knot slopes and enforce requested boundary conditions.
    knot_slopes = _normalize_knot_slopes(unnormalized_knot_slopes,
                                         min_knot_slope)
    if boundary_slopes == 'unconstrained':
      self._knot_slopes = knot_slopes
    elif boundary_slopes == 'lower_identity':
      ones = jnp.ones(pad_shape, self._dtype)
      self._knot_slopes = jnp.concatenate([ones, knot_slopes[..., 1:]], axis=-1)
    elif boundary_slopes == 'upper_identity':
      ones = jnp.ones(pad_shape, self._dtype)
      self._knot_slopes = jnp.concatenate(
          [knot_slopes[..., :-1], ones], axis=-1)
    elif boundary_slopes == 'identity':
      ones = jnp.ones(pad_shape, self._dtype)
      self._knot_slopes = jnp.concatenate(
          [ones, knot_slopes[..., 1:-1], ones], axis=-1)
    elif boundary_slopes == 'circular':
      self._knot_slopes = jnp.concatenate(
          [knot_slopes[..., :-1], knot_slopes[..., :1]], axis=-1)
    else:
      raise ValueError(f'Unknown option for boundary slopes:'
                       f' `{boundary_slopes}`.')

  @property
  def num_bins(self) -> int:
    """The number of segments on the interval."""
    return self._num_bins

  @property
  def knot_slopes(self) -> Array:
    """The slopes at the knot points."""
    return self._knot_slopes

  @property
  def x_pos(self) -> Array:
    """The bin boundaries on the `x`-axis."""
    return self._x_pos

  @property
  def y_pos(self) -> Array:
    """The bin boundaries on the `y`-axis."""
    return self._y_pos

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    fn = jnp.vectorize(
        _rational_quadratic_spline_fwd, signature='(),(n),(n),(n)->(),()')
    y, logdet = fn(x, self._x_pos, self._y_pos, self._knot_slopes)
    return y, logdet

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    fn = jnp.vectorize(
        _rational_quadratic_spline_inv, signature='(),(n),(n),(n)->(),()')
    x, logdet = fn(y, self._x_pos, self._y_pos, self._knot_slopes)
    return x, logdet



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `conversion.py`."""

import sys

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized

from distrax._src.bijectors.bijector import Bijector
from distrax._src.bijectors.rational_quadratic_spline import RationalQuadraticSpline
from distrax._src.bijectors.tanh import Tanh
from distrax._src.distributions.categorical import Categorical
from distrax._src.distributions.distribution import Distribution
from distrax._src.distributions.normal import Normal
from distrax._src.distributions.transformed import Transformed
from distrax._src.utils import conversion
import jax
from jax import config as jax_config
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfb = tfp.bijectors
tfd = tfp.distributions
FLAGS = flags.FLAGS
flags.DEFINE_bool('test_jax_enable_x64', False,
                  'Whether to enable double precision for tests.')


def setUpModule():
  if not FLAGS.is_parsed():
    FLAGS(sys.argv, known_only=True)
  if FLAGS['test_jax_enable_x64'].value:
    jax_config.update('jax_enable_x64', True)


class AsBijectorTest(parameterized.TestCase):

  def test_num_bins_attr_of_rational_quadratic_spline(self):
    num_bins = 4
    bijector = RationalQuadraticSpline(
        jnp.zeros((3 * num_bins + 1,)),
        range_min=0.,
        range_max=1.)
    wrapped_bijector = conversion.as_bijector(bijector)
    assert isinstance(wrapped_bijector, RationalQuadraticSpline)
    self.assertIs(wrapped_bijector, bijector)
    # Access the `num_bins` attribute of a wrapped RationalQuadraticSpline.
    np.testing.assert_equal(wrapped_bijector.num_bins, num_bins)

  def test_on_tfp_bijector(self):
    inputs = jnp.array([0., 1.])
    bijector = tfb.Exp()
    wrapped_bijector = conversion.as_bijector(bijector)
    assert isinstance(wrapped_bijector, Bijector)
    np.testing.assert_array_almost_equal(
        wrapped_bijector.forward(inputs),
        bijector.forward(inputs))


class AsDistributionTest(parameterized.TestCase):

  def test_loc_attr_of_normal(self):
    dist = Normal(loc=0., scale=1.)
    wrapped_dist = conversion.as_distribution(dist)
    assert isinstance(wrapped_dist, Normal)
    self.assertIs(wrapped_dist, dist)
    # Access the `loc` attribute of a wrapped Normal.
    np.testing.assert_almost_equal(wrapped_dist.loc, 0.)

  def test_num_categories_attr_of_categorical(self):
    dist = Categorical(logits=jnp.array([0., 0., 0.]))
    wrapped_dist = conversion.as_distribution(dist)
    assert isinstance(wrapped_dist, Categorical)
    self.assertIs(wrapped_dist, dist)
    # Access the `num_categories` attribute of a wrapped Categorical.
    np.testing.assert_equal(wrapped_dist.num_categories, 3)

  def test_attrs_of_transformed_distribution(self):
    dist = Transformed(Normal(loc=0., scale=1.), bijector=lambda x: x)
    wrapped_dist = conversion.as_distribution(dist)
    assert isinstance(wrapped_dist, Transformed)
    self.assertIs(wrapped_dist, dist)
    # Access the `distribution` attribute of a wrapped Transformed.
    assert isinstance(wrapped_dist.distribution, Normal)
    # Access the `loc` attribute of a transformed Normal within a wrapped
    # Transformed.
    np.testing.assert_almost_equal(wrapped_dist.distribution.loc, 0.)

  def test_on_tfp_distribution(self):
    dist = tfd.Normal(loc=0., scale=1.)
    wrapped_dist = conversion.as_distribution(dist)
    assert isinstance(wrapped_dist, tfd.Normal)
    assert isinstance(wrapped_dist, Distribution)
    # Access the `loc` attribute of a wrapped Normal.
    np.testing.assert_almost_equal(wrapped_dist.loc, 0.)


class ToTfpTest(parameterized.TestCase):

  def test_on_distrax_distribution(self):
    dist = Normal(loc=0., scale=1.)
    wrapped_dist = conversion.to_tfp(dist)
    assert isinstance(wrapped_dist, Normal)
    # Access the `loc` attribute of a wrapped Normal.
    np.testing.assert_almost_equal(wrapped_dist.loc, 0.)

  def test_on_distrax_bijector(self):
    bij = Tanh()
    wrapped_bij = conversion.to_tfp(bij)
    assert isinstance(wrapped_bij, Tanh)
    # Call the `forward` attribute of a wrapped Tanh.
    np.testing.assert_equal(
        wrapped_bij.forward(np.zeros(())), bij.forward(np.zeros(())))

  def test_on_tfp_distribution(self):
    dist = tfd.Normal(0., 1.)
    wrapped_dist = conversion.to_tfp(dist)
    self.assertIs(wrapped_dist, dist)

  def test_on_tfp_bijector(self):
    bij = tfb.Exp()
    wrapped_bij = conversion.to_tfp(bij)
    self.assertIs(wrapped_bij, bij)


class AsFloatArrayTest(parameterized.TestCase):

  @parameterized.parameters(0, 0.1)
  def test_on_valid_scalar(self, x):
    y = conversion.as_float_array(x)
    self.assertIsInstance(y, jnp.ndarray)
    self.assertEqual(
        y.dtype, jnp.float64 if jax.config.x64_enabled else jnp.float32)

  @parameterized.parameters(True, 1j)
  def test_on_invalid_scalar(self, x):
    with self.assertRaises(ValueError):
      conversion.as_float_array(x)

  @parameterized.parameters(
      float, jnp.float_, jnp.float16, jnp.float32, jnp.float64, jnp.bfloat16)
  def test_on_float_array(self, dtype):
    x = jnp.zeros([], dtype)
    y = conversion.as_float_array(x)
    self.assertIs(y, x)

  @parameterized.parameters(
      int, jnp.int_, jnp.int8, jnp.int16, jnp.int32, jnp.int64,
      jnp.uint8, jnp.uint16, jnp.uint32, jnp.uint64)
  def test_on_int_array(self, dtype):
    x = jnp.zeros([], dtype)
    y = conversion.as_float_array(x)
    self.assertIsInstance(y, jnp.ndarray)
    self.assertEqual(
        y.dtype, jnp.float64 if jax.config.x64_enabled else jnp.float32)

  @parameterized.parameters(
      bool, jnp.bool_, complex, jnp.complex_, jnp.complex64, jnp.complex128)
  def test_on_invalid_array(self, dtype):
    x = jnp.zeros([], dtype)
    with self.assertRaises(ValueError):
      conversion.as_float_array(x)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `mixture_same_family.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions.categorical import Categorical
from distrax._src.distributions.mixture_same_family import MixtureSameFamily
from distrax._src.distributions.mvn_diag import MultivariateNormalDiag
from distrax._src.distributions.normal import Normal
from distrax._src.utils import equivalence
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions


class TFPLogitsCategoricalTFPMultivariateComponents(equivalence.EquivalenceTest
                                                   ):
  """Class to test distrax mixture against tfp mixture.

  There are 4 methods to create categorical and components distributions (one
  each for tfp and distrax). Those are used to instantiate the input
  distributions for the mixtures to be tested against each other. By overloading
  these methods, different combinations can be tested.

  This class tests the case when using tfp distributions (categorical from
  logits) as inputs in the tfp and distrax mixture.
  """

  def _make_tfp_categorical(self, logits, probs):
    """Categorical distribution for tfp reference mixture."""
    del probs
    return tfd.Categorical(logits=logits)

  def _make_tfp_components(self, key_loc, key_scale):
    """Components distribution for tfp reference mixture."""
    components_shape = self.batch_shape + (self.num_components,) + (2,)
    return tfd.MultivariateNormalDiag(
        loc=jax.random.normal(key=key_loc, shape=components_shape),
        scale_diag=jax.random.uniform(key=key_scale, shape=components_shape)+.5)

  # Define functions to create input distributions for the Distrax mixture. This
  # class tests Distrax mixture using the same TFP input distributions in both
  # TFP and Distrax. Subclasses will use different combinations.
  _make_categorical = _make_tfp_categorical
  _make_components = _make_tfp_components

  def setUp(self):
    super().setUp()
    self._init_distr_cls(MixtureSameFamily)

    self.batch_shape = (5, 4)
    self.num_components = 3

    logits_shape = self.batch_shape + (self.num_components,)
    logits = jax.random.normal(key=jax.random.PRNGKey(42),
                               shape=logits_shape)
    probs = jax.nn.softmax(logits, axis=-1)
    self.mixture_dist = self._make_categorical(logits, probs)
    self.tfp_mixture_dist = self._make_tfp_categorical(logits, probs)

    key_loc, key_scale = jax.random.split(jax.random.PRNGKey(42))
    self.components_dist = self._make_components(key_loc, key_scale)
    self.tfp_components_dist = self._make_tfp_components(key_loc, key_scale)

  def test_event_shape(self):
    super()._test_event_shape(
        (),
        dist_kwargs={
            'mixture_distribution': self.mixture_dist,
            'components_distribution': self.components_dist
        },
        tfp_dist_kwargs={
            'mixture_distribution': self.tfp_mixture_dist,
            'components_distribution': self.tfp_components_dist
        },
    )

  def test_batch_shape(self):
    super()._test_batch_shape(
        (),
        dist_kwargs={
            'mixture_distribution': self.mixture_dist,
            'components_distribution': self.components_dist
        },
        tfp_dist_kwargs={
            'mixture_distribution': self.tfp_mixture_dist,
            'components_distribution': self.tfp_components_dist
        },
    )

  def test_invalid_parameters(self):
    logits_shape = (1,) + self.batch_shape + (self.num_components,)
    logits = jnp.ones(logits_shape, dtype=jnp.float32)
    probs = jax.nn.softmax(logits, axis=-1)
    key_loc, key_scale = jax.random.split(jax.random.PRNGKey(42))
    self._test_raises_error(dist_kwargs={
        'mixture_distribution': self._make_categorical(logits, probs),
        'components_distribution': self._make_components(key_loc, key_scale),
    })

  @chex.all_variants
  @parameterized.named_parameters(
      ('empty shape', ()),
      ('int shape', 10),
      ('2-tuple shape', (10, 20)),
  )
  def test_sample_shape(self, sample_shape):
    super()._test_sample_shape(
        (),
        dist_kwargs={
            'mixture_distribution': self.mixture_dist,
            'components_distribution': self.components_dist
        },
        tfp_dist_kwargs={
            'mixture_distribution': self.tfp_mixture_dist,
            'components_distribution': self.tfp_components_dist
        },
        sample_shape=sample_shape)

  @chex.all_variants
  def test_sample_dtype(self):
    dist = self.distrax_cls(
        mixture_distribution=self.mixture_dist,
        components_distribution=self.components_dist)
    samples = self.variant(dist.sample)(seed=self.key)
    self.assertEqual(dist.dtype, samples.dtype)
    self.assertEqual(dist.dtype, self.components_dist.dtype)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('empty shape', ()),
      ('int shape', 10),
      ('2-tuple shape', (10, 20)),
  )
  def test_sample_and_log_prob(self, sample_shape):
    super()._test_sample_and_log_prob(
        dist_args=(),
        dist_kwargs={
            'mixture_distribution': self.mixture_dist,
            'components_distribution': self.components_dist
        },
        tfp_dist_kwargs={
            'mixture_distribution': self.tfp_mixture_dist,
            'components_distribution': self.tfp_components_dist
        },
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=2e-3))

  # `pmap` must have at least one non-None value in `in_axes`.
  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('mean', 'mean'),
      ('variance', 'variance'),
      ('stddev', 'stddev'),
  )
  def test_method(self, function_string):
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs={
            'mixture_distribution': self.mixture_dist,
            'components_distribution': self.components_dist
        },
        tfp_dist_kwargs={
            'mixture_distribution': self.tfp_mixture_dist,
            'components_distribution': self.tfp_components_dist
        },
        assertion_fn=self.assertion_fn(rtol=2e-3))

  def test_jittable(self):
    super()._test_jittable(
        dist_kwargs={
            'mixture_distribution': self.mixture_dist,
            'components_distribution': self.components_dist},
        assertion_fn=self.assertion_fn(rtol=1e-3))


class TFPLogitsCategoricalTFPUnivariateComponents(
    TFPLogitsCategoricalTFPMultivariateComponents):

  def _make_tfp_components(self, key_loc, key_scale):
    components_shape = self.batch_shape + (self.num_components,)
    return tfd.Normal(
        loc=jax.random.normal(key=key_loc, shape=components_shape),
        scale=jax.random.uniform(key=key_scale, shape=components_shape)+0.5)

  _make_components = _make_tfp_components


# Enough to only test one type of components for `tfp.Categorical` with `probs`.
class TFPProbsCategoricalTFPMultivariateComponents(
    TFPLogitsCategoricalTFPMultivariateComponents):

  def _make_categorical(self, logits, probs):
    del logits
    return tfd.Categorical(probs=probs)


class DistraxLogitsCategoricalTFPMultivariateComponents(
    TFPLogitsCategoricalTFPMultivariateComponents):

  def _make_categorical(self, logits, probs):
    del probs
    return Categorical(logits=logits)


class DistraxProbsCategoricalTFPMultivariateComponents(
    TFPLogitsCategoricalTFPMultivariateComponents):

  def _make_categorical(self, logits, probs):
    del logits
    return Categorical(probs=probs)


class DistraxLogitsCategoricalTFPUnivariateComponents(
    TFPLogitsCategoricalTFPUnivariateComponents):

  def _make_categorical(self, logits, probs):
    del probs
    return Categorical(logits=logits)


class DistraxLogitsCategoricalDistraxMultivariateComponents(
    DistraxLogitsCategoricalTFPMultivariateComponents):

  def _make_components(self, key_loc, key_scale):
    components_shape = self.batch_shape + (self.num_components,) + (2,)
    return MultivariateNormalDiag(
        loc=jax.random.normal(key=key_loc, shape=components_shape),
        scale_diag=jax.random.uniform(key=key_scale, shape=components_shape) +
        0.5)


class DistraxLogitsCategoricalDistraxUnivariateComponents(
    DistraxLogitsCategoricalTFPUnivariateComponents):

  def _make_components(self, key_loc, key_scale):
    components_shape = self.batch_shape + (self.num_components,)
    return Normal(
        loc=jax.random.normal(key=key_loc, shape=components_shape),
        scale=jax.random.uniform(key=key_scale, shape=components_shape) + 0.5)


class TFPLogitsCategoricalDistraxMultivariateComponents(
    TFPLogitsCategoricalTFPMultivariateComponents):

  def _make_components(self, key_loc, key_scale):
    components_shape = self.batch_shape + (self.num_components,) + (2,)
    return MultivariateNormalDiag(
        loc=jax.random.normal(key=key_loc, shape=components_shape),
        scale_diag=jax.random.uniform(key=key_scale, shape=components_shape) +
        0.5)


class TFPLogitsCategoricalDistraxUnivariateComponents(
    TFPLogitsCategoricalTFPUnivariateComponents):

  def _make_components(self, key_loc, key_scale):
    components_shape = self.batch_shape + (self.num_components,)
    return Normal(
        loc=jax.random.normal(key=key_loc, shape=components_shape),
        scale=jax.random.uniform(key=key_scale, shape=components_shape) + 0.5)


class MixtureSameFamilySlicingTest(parameterized.TestCase):
  """Class to test the `getitem` method."""

  def setUp(self):
    super().setUp()
    self.loc = np.random.randn(2, 3, 4, 5)
    self.scale_diag = np.abs(np.random.randn(2, 3, 4, 5))
    self.components_dist = MultivariateNormalDiag(
        loc=self.loc, scale_diag=self.scale_diag)
    self.logits = np.random.randn(2, 3, 4)
    self.mixture_dist = Categorical(logits=self.logits)
    self.dist = MixtureSameFamily(self.mixture_dist, self.components_dist)

  def assertion_fn(self, rtol):
    return lambda x, y: np.testing.assert_allclose(x, y, rtol=rtol)

  @parameterized.named_parameters(
      ('single element', 1, (3,)),
      ('range', slice(-1), (1, 3)),
      ('range_2', (slice(None), slice(-1)), (2, 2)),
  )
  def test_slice(self, slice_, expected_batch_shape):
    sliced_dist = self.dist[slice_]
    self.assertEqual(sliced_dist.batch_shape, expected_batch_shape)
    self.assertEqual(sliced_dist.event_shape, self.dist.event_shape)
    self.assertEqual(sliced_dist.mixture_distribution.logits.shape[-1],
                     self.dist.mixture_distribution.logits.shape[-1])
    self.assertIsInstance(sliced_dist, MixtureSameFamily)
    self.assertIsInstance(
        sliced_dist.components_distribution, MultivariateNormalDiag)
    self.assertIsInstance(sliced_dist.mixture_distribution, Categorical)
    self.assertion_fn(rtol=2e-3)(
        sliced_dist.components_distribution.loc, self.loc[slice_])
    self.assertion_fn(rtol=2e-3)(
        sliced_dist.components_distribution.scale_diag, self.scale_diag[slice_])

  def test_slice_ellipsis(self):
    sliced_dist = self.dist[..., -1]
    expected_batch_shape = (2,)
    self.assertEqual(sliced_dist.batch_shape, expected_batch_shape)
    self.assertEqual(sliced_dist.event_shape, self.dist.event_shape)
    self.assertEqual(sliced_dist.mixture_distribution.logits.shape[-1],
                     self.dist.mixture_distribution.logits.shape[-1])
    self.assertIsInstance(sliced_dist, MixtureSameFamily)
    self.assertIsInstance(
        sliced_dist.components_distribution, MultivariateNormalDiag)
    self.assertIsInstance(sliced_dist.mixture_distribution, Categorical)
    self.assertion_fn(rtol=2e-3)(
        sliced_dist.components_distribution.loc,
        self.loc[:, -1, ...])
    self.assertion_fn(rtol=2e-3)(
        sliced_dist.components_distribution.scale_diag,
        self.scale_diag[:, -1, ...])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Independent distribution."""

from typing import Callable, Optional, Tuple, Union

import chex
from distrax._src.distributions import distribution as distrax_distribution
from distrax._src.utils import conversion
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions

Array = chex.Array
PRNGKey = chex.PRNGKey
DistributionLike = distrax_distribution.DistributionLike
EventT = distrax_distribution.EventT


class Independent(distrax_distribution.Distribution):
  """Independent distribution obtained from child distributions."""

  equiv_tfp_cls = tfd.Independent

  def __init__(self,
               distribution: DistributionLike,
               reinterpreted_batch_ndims: Optional[int] = None):
    """Initializes an Independent distribution.

    Args:
      distribution: Base distribution instance.
      reinterpreted_batch_ndims: Number of event dimensions.
    """
    super().__init__()
    distribution = conversion.as_distribution(distribution)
    self._distribution = distribution

    # Check if event shape is a tuple of integers (i.e. not nested).
    event_shape = distribution.event_shape
    if not (isinstance(event_shape, tuple) and
            all(isinstance(i, int) for i in event_shape)):
      raise ValueError(
          f"'Independent' currently only supports distributions with Array "
          f"events (i.e. not nested). Received '{distribution.name}' with "
          f"event shape '{distribution.event_shape}'.")

    dist_batch_shape = distribution.batch_shape
    if reinterpreted_batch_ndims is not None:
      dist_batch_ndims = len(dist_batch_shape)
      if reinterpreted_batch_ndims > dist_batch_ndims:
        raise ValueError(
            f'`reinterpreted_batch_ndims` is {reinterpreted_batch_ndims}, but'
            f' distribution `{distribution.name}` has only {dist_batch_ndims}'
            f' batch dimensions.')
      elif reinterpreted_batch_ndims < 0:
        raise ValueError(f'`reinterpreted_batch_ndims` can\'t be negative; got'
                         f' {reinterpreted_batch_ndims}.')
      self._reinterpreted_batch_ndims = reinterpreted_batch_ndims
    else:
      self._reinterpreted_batch_ndims = max(len(dist_batch_shape) - 1, 0)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    dist_batch_shape = self._distribution.batch_shape
    event_ndims = len(dist_batch_shape) - self._reinterpreted_batch_ndims
    return dist_batch_shape[event_ndims:] + self._distribution.event_shape

  @property
  def distribution(self):
    return self._distribution

  @property
  def reinterpreted_batch_ndims(self) -> int:
    return self._reinterpreted_batch_ndims

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    dist_batch_shape = self._distribution.batch_shape
    d = len(dist_batch_shape) - self.reinterpreted_batch_ndims
    return dist_batch_shape[:d]

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    return self._distribution.sample(seed=key, sample_shape=n)

  def _sample_n_and_log_prob(self, key: PRNGKey, n: int) -> Tuple[Array, Array]:
    """See `Distribution._sample_n_and_log_prob`."""
    samples, log_prob = self._distribution.sample_and_log_prob(
        seed=key, sample_shape=n)
    log_prob = self._reduce(jnp.sum, log_prob)
    return samples, log_prob

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    return self._reduce(jnp.sum, self._distribution.log_prob(value))

  def entropy(self) -> Array:
    """See `Distribution.entropy`."""
    return self._reduce(jnp.sum, self._distribution.entropy())

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return self._reduce(jnp.sum, self._distribution.log_cdf(value))

  def mean(self) -> Array:
    """Calculates the mean."""
    return self._distribution.mean()

  def median(self) -> Array:
    """Calculates the median."""
    return self._distribution.median()

  def variance(self) -> Array:
    """Calculates the variance."""
    return self._distribution.variance()

  def stddev(self) -> Array:
    """Calculates the standard deviation."""
    return self._distribution.stddev()

  def mode(self) -> Array:
    """Calculates the mode."""
    return self._distribution.mode()

  def _reduce(self, fn: Callable[..., Array], value: Array) -> Array:
    return fn(value,
              axis=[-i - 1 for i in range(0, self.reinterpreted_batch_ndims)])

  def __getitem__(self, index) -> 'Independent':
    """See `Distribution.__getitem__`."""
    index = distrax_distribution.to_batch_shape_index(self.batch_shape, index)
    return Independent(
        distribution=self.distribution[index],
        reinterpreted_batch_ndims=self.reinterpreted_batch_ndims)


def _kl_divergence_independent_independent(
    dist1: Union[Independent, tfd.Independent],
    dist2: Union[Independent, tfd.Independent],
    *args, **kwargs) -> Array:
  """Batched KL divergence `KL(dist1 || dist2)` for Independent distributions.

  Args:
    dist1: instance of an Independent distribution.
    dist2: instance of an Independent distribution.
    *args: Additional args.
    **kwargs: Additional kwargs.

  Returns:
    Batchwise `KL(dist1 || dist2)`
  """
  p = dist1.distribution
  q = dist2.distribution

  if dist1.event_shape == dist2.event_shape:
    if p.event_shape == q.event_shape:
      num_reduce_dims = len(dist1.event_shape) - len(p.event_shape)
      reduce_dims = [-i - 1 for i in range(0, num_reduce_dims)]
      kl_divergence = jnp.sum(p.kl_divergence(q, *args, **kwargs),
                              axis=reduce_dims)
    else:
      raise NotImplementedError(
          f'KL between Independents whose inner distributions have different '
          f'event shapes is not supported: obtained {p.event_shape} and '
          f'{q.event_shape}.')
  else:
    raise ValueError(f'Event shapes {dist1.event_shape} and {dist2.event_shape}'
                     f' do not match.')

  return kl_divergence


# Register the KL functions with TFP.
tfd.RegisterKL(Independent, Independent)(_kl_divergence_independent_independent)
tfd.RegisterKL(Independent, Independent.equiv_tfp_cls)(
    _kl_divergence_independent_independent)
tfd.RegisterKL(Independent.equiv_tfp_cls, Independent)(
    _kl_divergence_independent_independent)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `linear.py`."""

from absl.testing import absltest
from absl.testing import parameterized
from distrax._src.bijectors import linear
import jax.numpy as jnp


class MockLinear(linear.Linear):

  def forward_and_log_det(self, x):
    raise Exception  # pylint:disable=broad-exception-raised


class LinearTest(parameterized.TestCase):

  @parameterized.parameters(
      {'event_dims': 1, 'batch_shape': (), 'dtype': jnp.float16},
      {'event_dims': 10, 'batch_shape': (2, 3), 'dtype': jnp.float32})
  def test_properties(self, event_dims, batch_shape, dtype):
    bij = MockLinear(event_dims, batch_shape, dtype)
    self.assertEqual(bij.event_ndims_in, 1)
    self.assertEqual(bij.event_ndims_out, 1)
    self.assertTrue(bij.is_constant_jacobian)
    self.assertTrue(bij.is_constant_log_det)
    self.assertEqual(bij.event_dims, event_dims)
    self.assertEqual(bij.batch_shape, batch_shape)
    self.assertEqual(bij.dtype, dtype)
    with self.assertRaises(NotImplementedError):
      bij.matrix  # pylint: disable=pointless-statement


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Importance sampling."""

import chex
from distrax._src.distributions import distribution
import jax.numpy as jnp


Array = chex.Array
DistributionLike = distribution.DistributionLike


def importance_sampling_ratios(
    target_dist: DistributionLike,
    sampling_dist: DistributionLike,
    event: Array
) -> Array:
  """Compute importance sampling ratios given target and sampling distributions.

  Args:
    target_dist: Target probability distribution.
    sampling_dist: Sampling probability distribution.
    event: Samples.

  Returns:
    Importance sampling ratios.
  """
  log_pi_a = target_dist.log_prob(event)
  log_mu_a = sampling_dist.log_prob(event)
  rho = jnp.exp(log_pi_a - log_mu_a)
  return rho



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility functions for conversion between different types."""

from typing import Optional, Union

import chex
from distrax._src.bijectors import bijector
from distrax._src.bijectors import bijector_from_tfp
from distrax._src.bijectors import lambda_bijector
from distrax._src.bijectors import sigmoid
from distrax._src.bijectors import tanh
from distrax._src.bijectors import tfp_compatible_bijector
from distrax._src.distributions import distribution
from distrax._src.distributions import distribution_from_tfp
from distrax._src.distributions import tfp_compatible_distribution
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfb = tfp.bijectors
tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
BijectorLike = bijector.BijectorLike
DistributionLike = distribution.DistributionLike


def to_tfp(obj: Union[bijector.Bijector, tfb.Bijector,
                      distribution.Distribution, tfd.Distribution],
           name: Optional[str] = None):
  """Converts a distribution or bijector to a TFP-compatible equivalent object.

  The returned object is not necessarily of type `tfb.Bijector` or
  `tfd.Distribution`; rather, it is a Distrax object that implements TFP
  functionality so that it can be used in TFP.

  If the input is already of TFP type, it is returned unchanged.

  Args:
    obj: The distribution or bijector to be converted to TFP.
    name: The name of the resulting object.

  Returns:
    A TFP-compatible equivalent distribution or bijector.
  """
  if isinstance(obj, (tfb.Bijector, tfd.Distribution)):
    return obj
  elif isinstance(obj, bijector.Bijector):
    return tfp_compatible_bijector.tfp_compatible_bijector(obj, name)
  elif isinstance(obj, distribution.Distribution):
    return tfp_compatible_distribution.tfp_compatible_distribution(obj, name)
  else:
    raise TypeError(
        f"`to_tfp` can only convert objects of type: `distrax.Bijector`,"
        f" `tfb.Bijector`, `distrax.Distribution`, `tfd.Distribution`. Got type"
        f" `{type(obj)}`.")


def as_bijector(obj: BijectorLike) -> bijector.BijectorT:
  """Converts a bijector-like object to a Distrax bijector.

  Bijector-like objects are: Distrax bijectors, TFP bijectors, and callables.
  Distrax bijectors are returned unchanged. TFP bijectors are converted to a
  Distrax equivalent. Callables are wrapped by `distrax.Lambda`, with a few
  exceptions where an explicit implementation already exists and is returned.

  Args:
    obj: The bijector-like object to be converted.

  Returns:
    A Distrax bijector.
  """
  if isinstance(obj, bijector.Bijector):
    return obj
  elif isinstance(obj, tfb.Bijector):
    return bijector_from_tfp.BijectorFromTFP(obj)
  elif obj is jax.nn.sigmoid:
    return sigmoid.Sigmoid()
  elif obj is jnp.tanh:
    return tanh.Tanh()
  elif callable(obj):
    return lambda_bijector.Lambda(obj)
  else:
    raise TypeError(
        f"A bijector-like object can be a `distrax.Bijector`, a `tfb.Bijector`,"
        f" or a callable. Got type `{type(obj)}`.")


def as_distribution(obj: DistributionLike) -> distribution.DistributionT:
  """Converts a distribution-like object to a Distrax distribution.

  Distribution-like objects are: Distrax distributions and TFP distributions.
  Distrax distributions are returned unchanged. TFP distributions are converted
  to a Distrax equivalent.

  Args:
    obj: A distribution-like object to be converted.

  Returns:
    A Distrax distribution.
  """
  if isinstance(obj, distribution.Distribution):
    return obj
  elif isinstance(obj, tfd.Distribution):
    return distribution_from_tfp.distribution_from_tfp(obj)
  else:
    raise TypeError(
        f"A distribution-like object can be a `distrax.Distribution` or a"
        f" `tfd.Distribution`. Got type `{type(obj)}`.")


def as_float_array(x: Numeric) -> Array:
  """Converts input to an array with floating-point dtype.

  If the input is already an array with floating-point dtype, it is returned
  unchanged.

  Args:
    x: input to convert.

  Returns:
    An array with floating-point dtype.
  """
  if not isinstance(x, (jax.Array, np.ndarray)):
    x = jnp.asarray(x)

  if jnp.issubdtype(x.dtype, jnp.floating):
    return x
  elif jnp.issubdtype(x.dtype, jnp.integer):
    return x.astype(jnp.float_)
  else:
    raise ValueError(
        f"Expected either floating or integer dtype, got {x.dtype}.")



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `categorical_uniform.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import categorical_uniform
import jax
import jax.numpy as jnp
import numpy as np

_NAMED_PARAMETERS = (
    dict(
        testcase_name='scalar',
        low=0.,
        high=1.,
        logits=np.zeros((7,)),
        num_bins=7,
        target_event_shape=(),
        target_sample_shape=(),
        target_batch_shape=(),
        target_low=np.zeros(()),
        target_high=np.ones(()),
        target_logits=np.zeros((7,)),
        target_entropy=np.float64(0.0),
        target_mean=np.float32(0.5),
        target_variance=np.float32(1/12),
    ),
    dict(
        testcase_name='one_dimensional',
        low=np.zeros((2,)),
        high=np.ones((2,)),
        logits=np.zeros((2, 7)),
        num_bins=7,
        target_event_shape=(),
        target_sample_shape=(2,),
        target_batch_shape=(2,),
        target_low=np.zeros((2,)),
        target_high=np.ones((2,)),
        target_logits=np.zeros((2, 7)),
        target_entropy=np.full((2,), 0.0, dtype=np.float64),
        target_mean=np.full((2,), 0.5),
        target_variance=np.full((2,), 1/12),
    ),
    dict(
        testcase_name='two_dimensional',
        low=np.zeros((2, 3)),
        high=np.ones((2, 3)),
        logits=np.zeros((2, 3, 7)),
        num_bins=7,
        target_event_shape=(),
        target_sample_shape=(2, 3),
        target_batch_shape=(2, 3),
        target_low=np.zeros((2, 3)),
        target_high=np.ones((2, 3)),
        target_logits=np.zeros((2, 3, 7)),
        target_entropy=np.full((2, 3), 0.0, dtype=np.float64),
        target_mean=np.full((2, 3), 0.5),
        target_variance=np.full((2, 3), 1/12),
    ),
    dict(
        testcase_name='broadcasted_low',
        low=0.,
        high=np.ones((2, 3)),
        logits=np.zeros((2, 3, 7)),
        num_bins=7,
        target_event_shape=(),
        target_sample_shape=(2, 3),
        target_batch_shape=(2, 3),
        target_low=np.zeros((2, 3)),
        target_high=np.ones((2, 3)),
        target_logits=np.zeros((2, 3, 7)),
        target_entropy=np.full((2, 3), 0.0, dtype=np.float64),
        target_mean=np.full((2, 3), 0.5),
        target_variance=np.full((2, 3), 1/12),
    ),
    dict(
        testcase_name='broadcasted_high',
        low=np.zeros((2, 3)),
        high=1.,
        logits=np.zeros((2, 3, 7)),
        num_bins=7,
        target_event_shape=(),
        target_sample_shape=(2, 3),
        target_batch_shape=(2, 3),
        target_low=np.zeros((2, 3)),
        target_high=np.ones((2, 3)),
        target_logits=np.zeros((2, 3, 7)),
        target_entropy=np.full((2, 3), 0.0, dtype=np.float64),
        target_mean=np.full((2, 3), 0.5),
        target_variance=np.full((2, 3), 1/12),
    ),
    dict(
        testcase_name='broadcasted_logits',
        low=np.zeros((2, 3)),
        high=np.ones((2, 3)),
        logits=np.zeros((7,)),
        num_bins=7,
        target_event_shape=(),
        target_sample_shape=(2, 3),
        target_batch_shape=(2, 3),
        target_low=np.zeros((2, 3)),
        target_high=np.ones((2, 3)),
        target_logits=np.zeros((2, 3, 7)),
        target_entropy=np.full((2, 3), 0.0, dtype=np.float64),
        target_mean=np.full((2, 3), 0.5),
        target_variance=np.full((2, 3), 1/12),
    ),
)


class CategoricalUniformTest(parameterized.TestCase):

  def test_raises_on_wrong_logits(self):
    with self.assertRaises(ValueError):
      categorical_uniform.CategoricalUniform(
          low=0.0, high=1.0, logits=jnp.array(0.0)
      )

  @parameterized.named_parameters(*_NAMED_PARAMETERS)
  def test_batch_shape(self, *, low, high, logits, target_batch_shape, **_):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    self.assertEqual(distribution.batch_shape, target_batch_shape)

  @parameterized.named_parameters(*_NAMED_PARAMETERS)
  def test_event_shape(self, *, low, high, logits, target_event_shape, **_):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    self.assertEqual(distribution.event_shape, target_event_shape)

  @chex.all_variants
  @parameterized.named_parameters(*_NAMED_PARAMETERS)
  def test_sample_shape(self, *, low, high, logits, target_sample_shape, **_):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    sample = self.variant(distribution.sample)(seed=jax.random.PRNGKey(42))
    self.assertEqual(sample.shape, target_sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(*_NAMED_PARAMETERS)
  def test_sample_gradients(self, *, low, high, logits, **_):
    def summed_samples_fn(params):
      distribution = categorical_uniform.CategoricalUniform(
          low=params[0], high=params[1], logits=params[2])
      sample = distribution.sample(seed=jax.random.PRNGKey(42))
      return sample.sum()
    grad_fn = self.variant(jax.grad(summed_samples_fn))
    grad_low, grad_high, grad_logits = grad_fn(
        (jnp.float32(low), jnp.float32(high), jnp.float32(logits)))
    self.assertTrue(np.all(grad_low))  # Assert gradient is non-zero.
    self.assertTrue(np.all(grad_high))  # Assert gradient is non-zero.
    self.assertTrue(np.all(grad_logits))  # Assert gradient is non-zero.

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(*_NAMED_PARAMETERS)
  def test_entropy(self, *, low, high, logits, target_entropy, **_):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    chex.assert_trees_all_close(
        self.variant(distribution.entropy)(),
        target_entropy,
        atol=1e-4,
        rtol=1e-4,
    )

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(*_NAMED_PARAMETERS)
  def test_mean(self, *, low, high, logits, target_mean, **_):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    chex.assert_trees_all_close(
        self.variant(distribution.mean)(), target_mean, atol=1e-4, rtol=1e-4)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(*_NAMED_PARAMETERS)
  def test_variance(self, *, low, high, logits, target_variance, **_):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    chex.assert_trees_all_close(
        self.variant(distribution.variance)(),
        target_variance,
        atol=1e-4,
        rtol=1e-4,
    )

  @chex.all_variants
  @parameterized.named_parameters(*_NAMED_PARAMETERS)
  def test_log_prob(self, *, low, high, logits, target_sample_shape, **_):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    sample = jnp.full(target_sample_shape, 0.2, jnp.float32)
    log_prob = self.variant(distribution.log_prob)(sample)
    target_log_prob = jnp.zeros(target_sample_shape, jnp.float32)
    chex.assert_trees_all_close(log_prob, target_log_prob, atol=1e-4, rtol=1e-4)

  @parameterized.named_parameters(*_NAMED_PARAMETERS)
  def test_attributes(
      self, *,
      low, high, logits, target_low, target_high, target_logits, **_,
  ):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    with self.subTest('low'):
      chex.assert_trees_all_equal(distribution.low, target_low)
    with self.subTest('high'):
      chex.assert_trees_all_equal(distribution.high, target_high)
    with self.subTest('logits'):
      chex.assert_trees_all_equal(distribution.logits, target_logits)

  @parameterized.named_parameters(
      # Remove the scalar parameterization because slice would be out of range.
      named_parameters for named_parameters in _NAMED_PARAMETERS
      if named_parameters['testcase_name'] != 'scalar'
  )
  def test_slice(self, low, high, logits, **_):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    for name, key in (
        ('single_element', 1),
        ('range', slice(-1)),
        ('ellipsis', (Ellipsis, -1)),
    ):
      with self.subTest(name):
        chex.assert_trees_all_close(
            distribution[key].low,
            distribution.low[key],
            atol=1e-4,
            rtol=1e-4,
        )
        chex.assert_trees_all_close(
            distribution[key].high,
            distribution.high[key],
            atol=1e-4,
            rtol=1e-4,
        )
        dist_logits = distribution.logits
        chex.assert_trees_all_close(
            distribution[key].logits,
            dist_logits[key] if name != 'ellipsis' else dist_logits[..., -1, :],
            atol=1e-4,
            rtol=1e-4,
        )

  @chex.all_variants
  @parameterized.named_parameters(
      named_parameters for named_parameters in _NAMED_PARAMETERS
      if named_parameters['testcase_name'] == 'scalar'
  )
  def test_log_prob_outside_of_domain(
      self, *, low, high, logits, target_sample_shape, **_):
    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    with self.subTest('lower'):
      sample = jnp.full(target_sample_shape, -1, jnp.float32)
      log_prob = self.variant(distribution.log_prob)(sample)
      self.assertEqual(log_prob, -np.inf)
    with self.subTest('upper'):
      sample = jnp.full(target_sample_shape, +2, jnp.float32)
      log_prob = self.variant(distribution.log_prob)(sample)
      self.assertEqual(log_prob, -np.inf)

  @parameterized.named_parameters(
      named_parameters for named_parameters in _NAMED_PARAMETERS
      if named_parameters['testcase_name'] == 'two_dimensional'
  )
  def test_vmap_inputs(self, *, low, high, logits, target_sample_shape, **_):
    def log_prob_sum(distribution, sample):
      return distribution.log_prob(sample).sum()

    distribution = categorical_uniform.CategoricalUniform(
        low=low, high=high, logits=logits)
    sample = jnp.full(target_sample_shape, 0.2, jnp.float32)

    with self.subTest('no vmap'):
      actual = log_prob_sum(distribution, sample)
      expected = distribution.log_prob(sample).sum()
      chex.assert_trees_all_close(actual, expected, atol=1e-4, rtol=1e-4)

    with self.subTest('axis=0'):
      actual = jax.vmap(log_prob_sum, in_axes=0)(distribution, sample)
      expected = distribution.log_prob(sample).sum(axis=1)
      chex.assert_trees_all_close(actual, expected, atol=1e-4, rtol=1e-4)

    with self.subTest('axis=1'):
      actual = jax.vmap(log_prob_sum, in_axes=1)(distribution, sample)
      expected = distribution.log_prob(sample).sum(axis=0)
      chex.assert_trees_all_close(actual, expected, atol=1e-4, rtol=1e-4)

  @parameterized.named_parameters(
      named_parameters for named_parameters in _NAMED_PARAMETERS
      if named_parameters['testcase_name'] == 'two_dimensional'
  )
  def test_vmap_outputs(self, *, low, high, logits, target_sample_shape, **_):
    def summed_distribution(low, high, logits):
      return categorical_uniform.CategoricalUniform(
          low=low.sum(keepdims=True),
          high=high.sum(keepdims=True),
          logits=logits.sum(keepdims=True),
      )

    actual = jax.vmap(summed_distribution)(low, high, logits)
    expected = categorical_uniform.CategoricalUniform(
        low=low.sum(axis=1, keepdims=True),
        high=high.sum(axis=1, keepdims=True),
        logits=logits.sum(axis=1, keepdims=True),
    )
    np.testing.assert_equal(actual.batch_shape, expected.batch_shape)
    np.testing.assert_equal(actual.event_shape, expected.event_shape)
    sample = jnp.full(target_sample_shape, 0.2, jnp.float32)
    chex.assert_trees_all_close(
        actual.log_prob(sample),
        expected.log_prob(sample),
        atol=1e-4,
        rtol=1e-4,
    )

  def test_jitable(self):

    @jax.jit
    def jitted_function(event, dist):
      return dist.log_prob(event)

    dist = categorical_uniform.CategoricalUniform(
        low=0., high=1., logits=np.ones((7,)))
    event = dist.sample(seed=jax.random.PRNGKey(4242))
    log_prob = dist.log_prob(event)
    jitted_log_prob = jitted_function(event, dist)
    chex.assert_trees_all_close(jitted_log_prob, log_prob, atol=1e-4, rtol=1e-4)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `beta.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import beta as base_beta
from distrax._src.utils import equivalence
import jax.numpy as jnp
import numpy as np


class BetaTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(base_beta.Beta)

  @parameterized.named_parameters(
      ('0d params', (), (), ()),
      ('1d params', (2,), (2,), (2,)),
      ('2d params, no broadcast', (3, 2), (3, 2), (3, 2)),
      ('2d params, broadcasted alpha', (2,), (3, 2), (3, 2)),
      ('2d params, broadcasted beta', (3, 2), (2,), (3, 2)),
  )
  def test_properties(self, shape_of_alpha, shape_of_beta, batch_shape):
    rng = np.random.default_rng(42)
    alpha = 0.1 + rng.uniform(size=shape_of_alpha)
    beta = 0.1 + rng.uniform(size=shape_of_beta)
    dist = base_beta.Beta(alpha, beta)
    self.assertEqual(dist.event_shape, ())
    self.assertEqual(dist.batch_shape, batch_shape)
    self.assertion_fn(rtol=1e-3)(
        dist.alpha, np.broadcast_to(alpha, batch_shape))
    self.assertion_fn(rtol=1e-3)(dist.beta, np.broadcast_to(beta, batch_shape))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std beta, no shape', (1, 1), ()),
      ('1d std beta, int shape', (1, 1), 1),
      ('1d std beta, 1-tuple shape', (1, 1), (1,)),
      ('1d std beta, 2-tuple shape', (1, 1), (2, 2)),
      ('2d std beta, no shape', (np.ones(2), np.ones(2)), ()),
      ('2d std beta, int shape', ([1, 1], [1, 1]), 1),
      ('2d std beta, 1-tuple shape', (np.ones(2), np.ones(2)), (1,)),
      ('2d std beta, 2-tuple shape', ([1, 1], [1, 1]), (2, 2)),
      ('rank 2 std beta, 2-tuple shape', (np.ones((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted alpha', (1, np.ones(3)), (2, 2)),
      ('broadcasted beta', (np.ones(3), 1), ()),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_shape(distr_params, dict(), sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('sample, float16', 'sample', jnp.float16),
      ('sample, float32', 'sample', jnp.float32),
      ('sample_and_log_prob, float16', 'sample_and_log_prob', jnp.float16),
      ('sample_and_log_prob, float32', 'sample_and_log_prob', jnp.float32),
  )
  def test_sample_dtype(self, method, dtype):
    dist = self.distrax_cls(alpha=jnp.ones((), dtype), beta=jnp.ones((), dtype))
    samples = self.variant(getattr(dist, method))(seed=self.key)
    samples = samples[0] if method == 'sample_and_log_prob' else samples
    self.assertEqual(samples.dtype, dist.dtype)
    self.assertEqual(samples.dtype, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('sample', 'sample'),
      ('sample_and_log_prob', 'sample_and_log_prob'),
  )
  def test_sample_values(self, method):
    rng = np.random.default_rng(42)
    alpha = jnp.array(np.abs(rng.normal(size=(4, 3, 2))))
    beta = jnp.array(np.abs(rng.normal(size=(4, 3, 2))))
    n_samples = 100000
    dist = self.distrax_cls(alpha, beta)
    sample_fn = self.variant(
        lambda key: getattr(dist, method)(seed=key, sample_shape=n_samples))
    samples = sample_fn(self.key)
    samples = samples[0] if method == 'sample_and_log_prob' else samples
    self.assertEqual(samples.shape, (n_samples,) + (4, 3, 2))
    self.assertTrue(np.all(np.logical_and(samples >= 0., samples <= 1.)))
    self.assertion_fn(rtol=0.1)(np.mean(samples, axis=0), dist.mean())
    self.assertion_fn(rtol=0.1)(np.std(samples, axis=0), dist.stddev())

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std beta, no shape', (11, 3), ()),
      ('1d std beta, int shape', (11, 3), 1),
      ('1d std beta, 1-tuple shape', (11, 3), (1,)),
      ('1d std beta, 2-tuple shape', (1, 1), (2, 2)),
      ('2d std beta, no shape', (np.ones(2), np.ones(2)), ()),
      ('2d std beta, int shape', ([1, 1], [1, 1]), 1),
      ('2d std beta, 1-tuple shape', (np.ones(2), np.ones(2)), (1,)),
      ('2d std beta, 2-tuple shape', ([1, 1], [1, 1]), (2, 2)),
      ('rank 2 std beta, 2-tuple shape', (np.ones((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted alpha', (1, np.ones(3)), (2, 2)),
      ('broadcasted beta', (np.ones(3), 1), ()),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_and_log_prob(
        dist_args=distr_params,
        dist_kwargs=dict(),
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d dist, 1d value', (3.1, 1), 0.6),
      ('1d dist, 2d value', (0.5, 0.1), np.array([0.3, 0.8])),
      ('2d dist, 1d value', (0.5 + np.zeros(2), 0.3 * np.ones(2)), 0.7),
      ('2d broadcasted dist, 1d value', (0.4 + np.zeros(2), 0.8), 0.7),
      ('2d dist, 2d value',
       ([0.1, 0.5], 0.9 * np.ones(2)), np.array([0.2, 0.7])),
      ('edge cases alpha=1', (1., np.array([0.5, 2.])), np.array([0., 1.])),
      ('edge cases beta=1', (np.array([0.5, 2.]), 1.), np.array([0., 1.])),
  )
  def test_methods_with_value(self, distr_params, value):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    value = np.asarray(value, dtype=np.float32)
    for method in ['prob', 'log_prob', 'cdf', 'log_cdf', 'survival_function',
                   'log_survival_function']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=distr_params,
            call_args=(value,),
            assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('no broadcast', ([0.1, 1.3, 0.5], [0.5, 1.3, 1.5])),
      ('broadcasted alpha', (0.5, [0.5, 1.3, 1.5])),
      ('broadcasted beta', ([0.1, 1.3, 0.5], 0.8)),
  )
  def test_method(self, distr_params):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    for method in ['entropy', 'mean', 'variance', 'stddev']:
      with self.subTest(method=method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=distr_params,
            assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('alpha>1, beta>1, no broadcast', 1.5, 2.5, 0.25),
      ('alpha>1, beta>1, broadcasted alpha', 1.5, [2.5, 5.5], [0.25, 0.1]),
      ('alpha>1, beta>1, broadcasted beta', [1.5, 4.5], 2.5, [0.25, 0.7]),
      ('alpha<1, beta<1', 0.5, 0.1, np.nan),
      ('alpha=1, beta=1', 1., 1., np.nan),
      ('alpha=1, beta>1', 1., 1.5, 0.),
      ('alpha<1, beta>1', 0.5, 1.5, 0.),
      ('alpha>1, beta=1', 1.5, 1., 1.),
      ('alpha>1, beta<1', 1.5, 0.5, 1.),
  )
  def test_mode(self, alpha, beta, expected_result):
    dist = self.distrax_cls(alpha, beta)
    result = self.variant(dist.mode)()
    if np.any(np.isnan(expected_result)):
      self.assertTrue(jnp.isnan(result))
    else:
      self.assertion_fn(rtol=1e-3)(result, expected_result)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('distrax_to_distrax', 'distrax_to_distrax'),
      ('distrax_to_tfp', 'distrax_to_tfp'),
      ('tfp_to_distrax', 'tfp_to_distrax'),
  )
  def test_with_two_distributions(self, mode_string):
    rng = np.random.default_rng(42)
    alpha1 = np.abs(rng.normal(size=(4, 1, 2))).astype(np.float32)
    beta1 = np.abs(rng.normal(size=(4, 3, 2))).astype(np.float32)
    alpha2 = np.abs(rng.normal(size=(3, 2))).astype(np.float32)
    beta2 = np.abs(rng.normal(size=(3, 2))).astype(np.float32)
    for method in ['kl_divergence', 'cross_entropy']:
      with self.subTest(method=method):
        super()._test_with_two_distributions(
            attribute_string=method,
            mode_string=mode_string,
            dist1_kwargs={'alpha': alpha1, 'beta': beta1},
            dist2_kwargs={'alpha': alpha2, 'beta': beta2},
            tfp_dist1_kwargs={
                'concentration1': alpha1, 'concentration0': beta1},
            tfp_dist2_kwargs={
                'concentration1': alpha2, 'concentration0': beta2},
            assertion_fn=self.assertion_fn(rtol=3e-2))

  def test_jitable(self):
    super()._test_jittable(
        (0.1, 1.5), assertion_fn=self.assertion_fn(rtol=1e-3))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
      ('ellipsis', (Ellipsis, -1)),
  )
  def test_slice(self, slice_):
    rng = np.random.default_rng(42)
    alpha = jnp.array(np.abs(rng.normal(size=(4, 3, 2))))
    beta = jnp.array(np.abs(rng.normal(size=(4, 3, 2))))
    dist = self.distrax_cls(alpha, beta)
    self.assertion_fn(rtol=1e-3)(dist[slice_].alpha, alpha[slice_])
    self.assertion_fn(rtol=1e-3)(dist[slice_].beta, beta[slice_])

  def test_slice_different_parameterization(self):
    rng = np.random.default_rng(42)
    alpha = np.abs(rng.normal(size=(4, 3, 2)))
    beta = np.abs(rng.normal(size=(3, 2)))
    dist = self.distrax_cls(alpha, beta)
    self.assertion_fn(rtol=1e-3)(dist[0].alpha, alpha[0])
    self.assertion_fn(rtol=1e-3)(dist[0].beta, beta)  # Not slicing beta.


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Clipped distributions."""

from typing import Tuple

import chex
from distrax._src.distributions import distribution as base_distribution
from distrax._src.distributions import logistic
from distrax._src.distributions import normal
from distrax._src.utils import conversion
import jax.numpy as jnp


Array = chex.Array
PRNGKey = chex.PRNGKey
Numeric = chex.Numeric
DistributionLike = base_distribution.DistributionLike
EventT = base_distribution.EventT


class Clipped(base_distribution.Distribution):
  """A clipped distribution."""

  def __init__(
      self,
      distribution: DistributionLike,
      minimum: Numeric,
      maximum: Numeric):
    """Wraps a distribution clipping samples out of `[minimum, maximum]`.

    The samples outside of `[minimum, maximum]` are clipped to the boundary.
    The log probability of samples outside of this range is `-inf`.

    Args:
      distribution: a Distrax / TFP distribution to be wrapped.
      minimum: can be a `scalar` or `vector`; if a vector, must have fewer dims
        than `distribution.batch_shape` and must be broadcastable to it.
      maximum: can be a `scalar` or `vector`; if a vector, must have fewer dims
        than `distribution.batch_shape` and must be broadcastable to it.
    """
    super().__init__()
    if distribution.event_shape:
      raise ValueError('The wrapped distribution must have event shape ().')
    if (jnp.array(minimum).ndim > len(distribution.batch_shape) or
        jnp.array(maximum).ndim > len(distribution.batch_shape)):
      raise ValueError(
          'The minimum and maximum clipping boundaries must be scalars or'
          'vectors with fewer dimensions as the batch_shape of distribution:'
          'i.e. we can broadcast min/max to batch_shape but not viceversa.')
    self._distribution = conversion.as_distribution(distribution)
    self._minimum = jnp.broadcast_to(minimum, self._distribution.batch_shape)
    self._maximum = jnp.broadcast_to(maximum, self._distribution.batch_shape)
    self._log_prob_minimum = self._distribution.log_cdf(minimum)
    self._log_prob_maximum = self._distribution.log_survival_function(maximum)

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    raw_sample = self._distribution.sample(seed=key, sample_shape=[n])
    return jnp.clip(raw_sample, self._minimum, self._maximum)

  def _sample_n_and_log_prob(self, key: PRNGKey, n: int) -> Tuple[Array, Array]:
    """See `Distribution._sample_n_and_log_prob`."""
    samples = self._sample_n(key, n)
    return samples, self.log_prob(samples)

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    # The log_prob can be used to compute expectations by explicitly integrating
    # over the discrete and continuous elements.
    # Info about mixed distributions:
    # http://www.randomservices.org/random/dist/Mixed.html
    log_prob = jnp.where(
        jnp.equal(value, self._minimum),
        self._log_prob_minimum,
        jnp.where(jnp.equal(value, self._maximum),
                  self._log_prob_maximum,
                  self._distribution.log_prob(value)))
    # Giving -inf log_prob outside the boundaries.
    return jnp.where(
        jnp.logical_or(value < self._minimum, value > self._maximum),
        -jnp.inf,
        log_prob)

  @property
  def minimum(self) -> Array:
    return self._minimum

  @property
  def maximum(self) -> Array:
    return self._maximum

  @property
  def distribution(self) -> DistributionLike:
    return self._distribution

  @property
  def event_shape(self) -> Tuple[int, ...]:
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    return self._distribution.batch_shape

  def __getitem__(self, index) -> 'Clipped':
    """See `Distribution.__getitem__`."""
    index = base_distribution.to_batch_shape_index(self.batch_shape, index)
    return Clipped(
        distribution=self.distribution[index],
        minimum=self.minimum[index],
        maximum=self.maximum[index])


class ClippedNormal(Clipped):
  """A clipped normal distribution."""

  def __init__(
      self, loc: Numeric, scale: Numeric, minimum: Numeric, maximum: Numeric):
    distribution = normal.Normal(loc=loc, scale=scale)
    super().__init__(distribution, minimum=minimum, maximum=maximum)


class ClippedLogistic(Clipped):
  """A clipped logistic distribution."""

  def __init__(
      self, loc: Numeric, scale: Numeric, minimum: Numeric, maximum: Numeric):
    distribution = logistic.Logistic(loc=loc, scale=scale)
    super().__init__(distribution, minimum=minimum, maximum=maximum)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `transformed.py`."""

import functools

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors import block
from distrax._src.bijectors import lambda_bijector
from distrax._src.bijectors import masked_coupling
from distrax._src.bijectors import scalar_affine
from distrax._src.bijectors import sigmoid
from distrax._src.distributions import normal
from distrax._src.distributions import transformed
from distrax._src.utils import conversion
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions
tfb = tfp.bijectors


def _with_additional_parameters(params, all_named_parameters):
  """Convenience function for appending a cartesian product of parameters."""
  for name, param in params:
    for named_params in all_named_parameters:
      yield (f'{named_params[0]}; {name}',) + named_params[1:] + (param,)


def _with_base_dists(*all_named_parameters):
  """Partial of _with_additional_parameters to specify distrax and tfp base."""
  base_dists = (
      ('tfp_base', tfd.Normal),
      ('distrax_base', normal.Normal),
  )
  return _with_additional_parameters(base_dists, all_named_parameters)


class TransformedTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.seed = jax.random.PRNGKey(1234)

  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal', 0, 1),
      ('2d std normal', np.zeros(2), np.ones(2)),
      ('broadcasted loc', 0, np.ones(3)),
      ('broadcasted scale', np.ones(3), 1),
  ))
  def test_event_shape(self, mu, sigma, base_dist):
    base = base_dist(mu, sigma)
    bijector = tfb.Scale(2)
    dist = transformed.Transformed(base, bijector)
    tfp_dist = tfd.TransformedDistribution(conversion.to_tfp(base), bijector)

    assert dist.event_shape == tfp_dist.event_shape

  @parameterized.named_parameters(
      ('tfp_normal, tfp_scale',
       lambda: tfd.Normal(0, 1), lambda: tfb.Scale(2)),
      ('tfp_normal, tfp_shift',
       lambda: tfd.Normal(0, 1), lambda: tfb.Shift(3.0)),
      ('tfp_normal, tfp_tanh',
       lambda: tfd.Normal(0, 1), tfb.Tanh),
  )
  def test_dtype_is_consistent_with_tfp(self, dist_fn, bijector_fn):
    base = dist_fn()
    bijector = bijector_fn()
    dist = transformed.Transformed(base, bijector)
    tfp_dist = tfd.TransformedDistribution(conversion.to_tfp(base), bijector)
    assert dist.dtype == tfp_dist.dtype

  @chex.all_variants
  @parameterized.named_parameters(
      ('tfp_normal, dx_lambda_scale', lambda: tfd.Normal(0, 1),
       lambda: lambda x: x * 2, jnp.float32),
      ('tfp_normal, dx_lambda_shift', lambda: tfd.Normal(0, 1),
       lambda: lambda x: x + 3.0, jnp.float32),
      ('tfp_normal, dx_lambda_tanh', lambda: tfd.Normal(0, 1),
       lambda: jnp.tanh, jnp.float32),
  )
  def test_dtype_is_as_expected(self, dist_fn, bijector_fn, expected_dtype):
    base = dist_fn()
    bijector = bijector_fn()
    dist = transformed.Transformed(base, bijector)
    sample = self.variant(dist.sample)(seed=self.seed)
    assert dist.dtype == sample.dtype
    assert dist.dtype == expected_dtype

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal, no shape', 0, 1, ()),
      ('1d std normal, int shape', 0, 1, 1),
      ('1d std normal, 1-tuple shape', 0, 1, (1,)),
      ('1d std normal, 2-tuple shape', 0, 1, (2, 2)),
      ('2d std normal, no shape', np.zeros(2), np.ones(2), ()),
      ('2d std normal, int shape', [0, 0], [1, 1], 1),
      ('2d std normal, 1-tuple shape', np.zeros(2), np.ones(2), (1,)),
      ('2d std normal, 2-tuple shape', [0, 0], [1, 1], (2, 2)),
      ('rank 2 std normal, 2-tuple shape', np.zeros(
          (3, 2)), np.ones((3, 2)), (2, 2)),
      ('broadcasted loc', 0, np.ones(3), (2, 2)),
      ('broadcasted scale', np.ones(3), 1, ()),
  ))
  def test_sample_shape(self, mu, sigma, sample_shape, base_dist):
    base = base_dist(mu, sigma)
    bijector = tfb.Scale(2)
    dist = transformed.Transformed(base, bijector)
    def sample_fn(seed, sample_shape):
      return dist.sample(seed=seed, sample_shape=sample_shape)
    samples = self.variant(sample_fn, ignore_argnums=(1,), static_argnums=1)(
        self.seed, sample_shape)

    tfp_dist = tfd.TransformedDistribution(conversion.to_tfp(base), bijector)
    tfp_samples = tfp_dist.sample(sample_shape=sample_shape, seed=self.seed)

    chex.assert_equal_shape([samples, tfp_samples])

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d dist, 1d value', 0, 1, 1.),
      ('1d dist, 2d value', 0., 1., np.array([1., 2.])),
      ('2d dist, 1d value', np.zeros(2), np.ones(2), 1.),
      ('2d broadcasted dist, 1d value', np.zeros(2), 1, 1.),
      ('2d dist, 2d value', np.zeros(2), np.ones(2), np.array([1., 2.])),
      ('1d dist, 1d value, edge case', 0, 1, 200.),
  ))
  def test_log_prob(self, mu, sigma, value, base_dist):
    base = base_dist(mu, sigma)
    bijector = tfb.Scale(2)
    dist = transformed.Transformed(base, bijector)
    actual = self.variant(dist.log_prob)(value)

    tfp_dist = tfd.TransformedDistribution(conversion.to_tfp(base), bijector)
    expected = tfp_dist.log_prob(value)
    np.testing.assert_array_equal(actual, expected)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d dist, 1d value', 0, 1, 1.),
      ('1d dist, 2d value', 0., 1., np.array([1., 2.])),
      ('2d dist, 1d value', np.zeros(2), np.ones(2), 1.),
      ('2d broadcasted dist, 1d value', np.zeros(2), 1, 1.),
      ('2d dist, 2d value', np.zeros(2), np.ones(2), np.array([1., 2.])),
      ('1d dist, 1d value, edge case', 0, 1, 200.),
  ))
  def test_prob(self, mu, sigma, value, base_dist):
    base = base_dist(mu, sigma)
    bijector = tfb.Scale(2)
    dist = transformed.Transformed(base, bijector)
    actual = self.variant(dist.prob)(value)

    tfp_dist = tfd.TransformedDistribution(conversion.to_tfp(base), bijector)
    expected = tfp_dist.prob(value)
    np.testing.assert_array_equal(actual, expected)

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('1d std normal, no shape', 0, 1, ()),
      ('1d std normal, int shape', 0, 1, 1),
      ('1d std normal, 1-tuple shape', 0, 1, (1,)),
      ('1d std normal, 2-tuple shape', 0, 1, (2, 2)),
      ('2d std normal, no shape', np.zeros(2), np.ones(2), ()),
      ('2d std normal, int shape', [0, 0], [1, 1], 1),
      ('2d std normal, 1-tuple shape', np.zeros(2), np.ones(2), (1,)),
      ('2d std normal, 2-tuple shape', [0, 0], [1, 1], (2, 2)),
      ('rank 2 std normal, 2-tuple shape', np.zeros(
          (3, 2)), np.ones((3, 2)), (2, 2)),
      ('broadcasted loc', 0, np.ones(3), (2, 2)),
      ('broadcasted scale', np.ones(3), 1, ()),
  ))
  def test_sample_and_log_prob(self, mu, sigma, sample_shape, base_dist):
    base = base_dist(mu, sigma)
    bijector = tfb.Tanh()
    dist = transformed.Transformed(base, bijector)
    def sample_and_log_prob_fn(seed, sample_shape):
      return dist.sample_and_log_prob(seed=seed, sample_shape=sample_shape)
    samples, log_prob = self.variant(
        sample_and_log_prob_fn, ignore_argnums=(1,), static_argnums=(1,))(
            self.seed, sample_shape)
    expected_samples = bijector.forward(
        base.sample(seed=self.seed, sample_shape=sample_shape))

    tfp_dist = tfd.TransformedDistribution(conversion.to_tfp(base), bijector)
    tfp_samples = tfp_dist.sample(seed=self.seed, sample_shape=sample_shape)
    tfp_log_prob = tfp_dist.log_prob(samples)

    chex.assert_equal_shape([samples, tfp_samples])
    np.testing.assert_allclose(log_prob, tfp_log_prob, rtol=1e-2)
    np.testing.assert_allclose(samples, expected_samples, rtol=1e-2)

  @parameterized.named_parameters(
      ('1d-batched bijector,  unbatched sample', (2,), ()),
      ('1d-batched bijector, 1d-batched sample', (2,), (4,)),
      ('1d-batched bijector, 2d-batched sample', (2,), (4, 5)),
      ('2d-batched bijector,  unbatched sample', (5, 2), ()),
      ('2d-batched bijector, 1d-batched sample', (5, 2), (4,)),
      ('2d-batched bijector, 2d-batched sample', (5, 2), (4, 5)),
      ('3d-batched bijector,  unbatched sample', (7, 5, 2), ()),
      ('3d-batched bijector, 1d-batched sample', (7, 5, 2), (4,)),
      ('3d-batched bijector, 2d-batched sample', (7, 5, 2), (4, 5)),
  )
  def test_batched_bijector_shapes(self, batch_shape, sample_shape):
    base = tfd.MultivariateNormalDiag(jnp.zeros(3), jnp.ones(3))
    bijector = block.Block(tfb.Scale(jnp.ones(batch_shape + (3,))), 1)
    dist = transformed.Transformed(base, bijector)

    with self.subTest('batch_shape'):
      chex.assert_equal(dist.batch_shape, batch_shape)

    with self.subTest('sample.shape'):
      sample = dist.sample(seed=self.seed, sample_shape=sample_shape)
      chex.assert_equal(sample.shape, sample_shape + batch_shape + (3,))

    with self.subTest('sample_and_log_prob sample.shape'):
      sample, log_prob = dist.sample_and_log_prob(
          seed=self.seed, sample_shape=sample_shape)
      chex.assert_equal(sample.shape, sample_shape + batch_shape + (3,))

    with self.subTest('sample_and_log_prob log_prob.shape'):
      sample, log_prob = dist.sample_and_log_prob(
          seed=self.seed, sample_shape=sample_shape)
      chex.assert_equal(log_prob.shape, sample_shape + batch_shape)

    with self.subTest('sample_and_log_prob log_prob value'):
      sample, log_prob = dist.sample_and_log_prob(
          seed=self.seed, sample_shape=sample_shape)
      np.testing.assert_allclose(log_prob, dist.log_prob(sample))

  @chex.all_variants
  @parameterized.named_parameters(
      ('Scale-scalar-unbatched', tfb.Scale, 1, (), 3),
      ('Scale-scalar-batched', tfb.Scale, 1, (), (2, 3)),
      ('Scale-vector-unbatched', tfb.Scale, 1, 3, 3),
      ('Scale-vector-batched', tfb.Scale, 1, 3, (2, 3)),
      ('Scale-batched-unbatched', tfb.Scale, 1, (2, 3), 3),
      ('Scale-batched-batched', tfb.Scale, 1, (2, 3), (2, 3)),
      ('Matvec-vector-unbatched', tfb.ScaleMatvecDiag, 0, 3, 3),
      ('Matvec-vector-batched', tfb.ScaleMatvecDiag, 0, 3, (2, 3)),
      ('Matvec-batched-unbatched', tfb.ScaleMatvecDiag, 0, (2, 3), 3),
      ('Matvec-batched-batched', tfb.ScaleMatvecDiag, 0, (2, 3), (2, 3)),
  )
  def test_batched_bijector_against_tfp(
      self, bijector_fn, block_ndims, bijector_shape, params_shape):

    base = tfd.MultivariateNormalDiag(
        jnp.zeros(params_shape), jnp.ones(params_shape))

    tfp_bijector = bijector_fn(jnp.ones(bijector_shape))
    dx_bijector = block.Block(tfp_bijector, block_ndims)

    dx_dist = transformed.Transformed(base, dx_bijector)
    tfp_dist = tfd.TransformedDistribution(
        conversion.to_tfp(base), tfp_bijector)

    with self.subTest('event_shape property matches TFP'):
      np.testing.assert_equal(dx_dist.event_shape, tfp_dist.event_shape)

    with self.subTest('sample shape matches TFP'):
      dx_sample = self.variant(dx_dist.sample)(seed=self.seed)
      tfp_sample = self.variant(tfp_dist.sample)(seed=self.seed)
      chex.assert_equal_shape([dx_sample, tfp_sample])

    with self.subTest('log_prob(dx_sample) matches TFP'):
      dx_logp_dx = self.variant(dx_dist.log_prob)(dx_sample)
      tfp_logp_dx = self.variant(tfp_dist.log_prob)(dx_sample)
      np.testing.assert_allclose(dx_logp_dx, tfp_logp_dx, rtol=1e-2)

    with self.subTest('log_prob(tfp_sample) matches TFP'):
      dx_logp_tfp = self.variant(dx_dist.log_prob)(tfp_sample)
      tfp_logp_tfp = self.variant(tfp_dist.log_prob)(tfp_sample)
      np.testing.assert_allclose(dx_logp_tfp, tfp_logp_tfp, rtol=1e-2)

    with self.subTest('sample/lp shape is self-consistent'):
      second_sample, log_prob = self.variant(dx_dist.sample_and_log_prob)(
          seed=self.seed)
      chex.assert_equal_shape([dx_sample, second_sample])
      chex.assert_equal_shape([dx_logp_dx, log_prob])

  # These should all fail because the bijector's event_ndims is incompatible
  # with the base distribution's event_shape.
  @parameterized.named_parameters(
      ('scalar', 0),
      ('matrix', 2),
      ('3-tensor', 3),
  )
  def test_raises_on_incorrect_shape(self, block_dims):
    base = tfd.MultivariateNormalDiag(jnp.zeros((2, 3)), jnp.ones((2, 3)))
    scalar_bijector = tfb.Scale(jnp.ones((1, 2, 3)))
    block_bijector = block.Block(scalar_bijector, block_dims)
    with self.assertRaises(ValueError):
      transformed.Transformed(base, block_bijector)

  @chex.all_variants
  def test_bijector_that_assumes_batch_dimensions(self):
    # Create a Haiku conditioner that assumes a single batch dimension.
    def forward(x):
      network = hk.Sequential([hk.Flatten(preserve_dims=1), hk.Linear(3)])
      return network(x)
    init, apply = hk.transform(forward)
    params = init(self.seed, jnp.ones((2, 3)))
    conditioner = functools.partial(apply, params, self.seed)

    bijector = masked_coupling.MaskedCoupling(
        jnp.ones(3) > 0, conditioner, tfb.Scale)

    base = tfd.MultivariateNormalDiag(jnp.zeros((2, 3)), jnp.ones((2, 3)))
    dist = transformed.Transformed(base, bijector)
    # Exercise the trace-based functions
    assert dist.batch_shape == (2,)
    assert dist.event_shape == (3,)
    assert dist.dtype == jnp.float32
    sample = self.variant(dist.sample)(seed=self.seed)
    assert sample.dtype == dist.dtype
    self.variant(dist.log_prob)(sample)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(_with_base_dists(
      ('entropy', 'entropy', 0., 1.),
      ('mean', 'mean', 0, 1),
      ('mean from list params', 'mean', [-1, 1], [1, 2]),
      ('mode', 'mode', 0, 1),
  ))
  def test_method(self, function_string, mu, sigma, base_dist):
    base = base_dist(mu, sigma)
    bijector = tfb.Scale(2)
    dist = transformed.Transformed(base, bijector)
    tfp_dist = tfd.TransformedDistribution(conversion.to_tfp(base), bijector)

    np.testing.assert_allclose(
        self.variant(getattr(dist, function_string))(),
        getattr(tfp_dist, function_string)())

  @chex.all_variants
  @parameterized.named_parameters(_with_base_dists(
      ('int16', np.array([0, 0], dtype=np.int16)),
      ('int32', np.array([0, 0], dtype=np.int32)),
      ('int64', np.array([0, 0], dtype=np.int64)),
  ))
  def test_integer_inputs(self, inputs, base_dist):
    base = base_dist(jnp.zeros_like(inputs, dtype=jnp.float32),
                     jnp.ones_like(inputs, dtype=jnp.float32))
    bijector = scalar_affine.ScalarAffine(shift=0.0)
    dist = transformed.Transformed(base, bijector)

    log_prob = self.variant(dist.log_prob)(inputs)

    standard_normal_log_prob_of_zero = -0.9189385
    expected_log_prob = jnp.full_like(
        inputs, standard_normal_log_prob_of_zero, dtype=jnp.float32)

    np.testing.assert_array_equal(log_prob, expected_log_prob)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'tfp_to_distrax'))
  def test_kl_divergence(self, mode_string):
    base_dist1 = tfd.Normal([0.1, 0.5, 0.9], [0.1, 1.1, 2.5])
    base_dist2 = tfd.Normal(-0.1, 1.5)
    bij_tfp1 = tfb.Identity()
    bij_tfp2 = tfb.Identity()
    bij_distrax1 = bij_tfp1
    bij_distrax2 = lambda_bijector.Lambda(lambda x: x)
    tfp_dist1 = tfd.TransformedDistribution(base_dist1, bij_tfp1)
    tfp_dist2 = tfd.TransformedDistribution(base_dist2, bij_tfp2)
    distrax_dist1 = transformed.Transformed(base_dist1, bij_distrax1)
    distrax_dist2 = transformed.Transformed(base_dist2, bij_distrax2)

    expected_result_fwd = base_dist1.kl_divergence(base_dist2)
    expected_result_inv = base_dist2.kl_divergence(base_dist1)

    distrax_fn1 = self.variant(distrax_dist1.kl_divergence)
    distrax_fn2 = self.variant(distrax_dist2.kl_divergence)

    if mode_string == 'distrax_to_distrax':
      result_fwd = distrax_fn1(distrax_dist2)
      result_inv = distrax_fn2(distrax_dist1)
    elif mode_string == 'distrax_to_tfp':
      result_fwd = distrax_fn1(tfp_dist2)
      result_inv = distrax_fn2(tfp_dist1)
    elif mode_string == 'tfp_to_distrax':
      result_fwd = tfp_dist1.kl_divergence(distrax_dist2)
      result_inv = tfp_dist2.kl_divergence(distrax_dist1)
    else:
      raise ValueError(f'Unsupported mode: {mode_string}')
    np.testing.assert_allclose(result_fwd, expected_result_fwd, rtol=1e-2)
    np.testing.assert_allclose(result_inv, expected_result_inv, rtol=1e-2)

  @chex.all_variants(with_pmap=False)
  def test_kl_divergence_on_same_instance_of_distrax_bijector(self):
    base_dist1 = tfd.Normal([0.1, 0.5, 0.9], [0.1, 1.1, 2.5])
    base_dist2 = tfd.Normal(-0.1, 1.5)
    bij_distrax = sigmoid.Sigmoid()
    distrax_dist1 = transformed.Transformed(base_dist1, bij_distrax)
    distrax_dist2 = transformed.Transformed(base_dist2, bij_distrax)
    expected_result_fwd = base_dist1.kl_divergence(base_dist2)
    expected_result_inv = base_dist2.kl_divergence(base_dist1)
    result_fwd = self.variant(distrax_dist1.kl_divergence)(distrax_dist2)
    result_inv = self.variant(distrax_dist2.kl_divergence)(distrax_dist1)
    np.testing.assert_allclose(result_fwd, expected_result_fwd, rtol=1e-2)
    np.testing.assert_allclose(result_inv, expected_result_inv, rtol=1e-2)

  def test_kl_divergence_raises_on_event_shape(self):
    base_dist1 = tfd.MultivariateNormalDiag([0.1, 0.5, 0.9], [0.1, 1.1, 2.5])
    base_dist2 = tfd.Normal(-0.1, 1.5)
    bij1 = block.Block(lambda_bijector.Lambda(lambda x: x), ndims=1)
    bij2 = lambda_bijector.Lambda(lambda x: x)
    distrax_dist1 = transformed.Transformed(base_dist1, bij1)
    distrax_dist2 = transformed.Transformed(base_dist2, bij2)
    with self.assertRaises(ValueError):
      distrax_dist1.kl_divergence(distrax_dist2)

  def test_kl_divergence_raises_on_different_bijectors(self):
    base_dist1 = tfd.Normal([0.1, 0.5, 0.9], [0.1, 1.1, 2.5])
    base_dist2 = tfd.Normal(-0.1, 1.5)
    bij1 = lambda_bijector.Lambda(lambda x: x)
    bij2 = sigmoid.Sigmoid()
    distrax_dist1 = transformed.Transformed(base_dist1, bij1)
    distrax_dist2 = transformed.Transformed(base_dist2, bij2)
    with self.assertRaises(NotImplementedError):
      distrax_dist1.kl_divergence(distrax_dist2)

  def test_jittable(self):
    @jax.jit
    def f(x, d):
      return d.log_prob(x)

    base = normal.Normal(0, 1)
    bijector = scalar_affine.ScalarAffine(0, 1)
    dist = transformed.Transformed(base, bijector)
    x = np.zeros(())
    f(x, dist)

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper to adapt a Distrax bijector for use in TFP."""

from typing import Any, Optional, Tuple

import chex
from distrax._src.bijectors import bijector
from distrax._src.utils import math
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfb = tfp.bijectors
tfd = tfp.distributions

Array = chex.Array
Bijector = bijector.Bijector
TangentSpace = tfp.experimental.tangent_spaces.TangentSpace


def tfp_compatible_bijector(
    base_bijector: Bijector,
    name: Optional[str] = None):
  """Create a TFP-compatible bijector from a Distrax bijector.

  Given a Distrax bijector, return a wrapped bijector that behaves as a TFP
  bijector, to be used in TFP meta-bijectors and the TransformedDistribution.
  In particular, the wrapped bijector implements the methods
  `[forward|inverse]_event_ndims`, `[forward|inverse]_event_shape`,
  `[forward|inverse]_event_shape_tensor`, `[forward|inverse]_log_det_jacobian`,
  and the properties `[forward|inverse]_min_event_ndims`. Other attributes are
  delegated to the `base_bijector`.

  The methods of the resulting object do not take a `name` argument,
  unlike their TFP equivalents.

  The `shape` methods are implemented by tracing the `forward` and `inverse`
  methods of the bijector, applied to a zero tensor of the requested dtype. If
  the `forward` or `inverse` methods are not traceable or cannot be applied to a
  zero tensor, then we cannot guarantee the correctness of the result.

  Args:
    base_bijector: A Distrax bijector.
    name: The bijector name.

  Returns:
    An object that behaves like a TFP bijector.
  """

  name_ = name

  class TFPCompatibleBijector(base_bijector.__class__):
    """Class to wrap a Distrax bijector."""

    def __init__(self):
      self._is_injective = True
      self._is_permutation = False
      self._parts_interact = False

      self.dtype = None
      self.has_static_min_event_ndims = True
      self.forward_min_event_ndims = base_bijector.event_ndims_in
      self.inverse_min_event_ndims = base_bijector.event_ndims_out

    def __getattr__(self, name: str):
      return getattr(base_bijector, name)

    def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
      """See `Bijector.forward_and_log_det`."""
      return base_bijector.forward_and_log_det(x)

    @property
    def name(self) -> str:
      """The name of the wrapped bijector."""
      return name_ or f"TFPCompatible{base_bijector.name}"

    def experimental_batch_shape(self, x_event_ndims=None, y_event_ndims=None):
      raise NotImplementedError()

    def experimental_batch_shape_tensor(
        self, x_event_ndims=None, y_event_ndims=None):
      raise NotImplementedError()

    def forward_dtype(self, _: jnp.dtype) -> None:
      """Returns None, making no promise regarding dtypes."""
      return None

    def inverse_dtype(self, _: jnp.dtype) -> None:
      """Returns None, making no promise regarding dtypes."""
      return None

    def forward_event_ndims(self, event_ndims: int) -> int:
      """Returns the number of event dimensions of the output of `forward`."""
      extra_event_ndims = self._check_ndims(
          "Forward", event_ndims, base_bijector.event_ndims_in)
      return base_bijector.event_ndims_out + extra_event_ndims

    def inverse_event_ndims(self, event_ndims: int) -> int:
      """Returns the number of event dimensions of the output of `inverse`."""
      extra_event_ndims = self._check_ndims(
          "Inverse", event_ndims, base_bijector.event_ndims_out)
      return base_bijector.event_ndims_in + extra_event_ndims

    def forward_event_shape(self, event_shape) -> tfp.tf2jax.TensorShape:
      """Returns the shape of the output of `forward` as a `TensorShape`."""
      self._check_shape("Forward", event_shape, base_bijector.event_ndims_in)
      forward_event_shape = jax.eval_shape(
          base_bijector.forward, jnp.zeros(event_shape)).shape
      return tfp.tf2jax.TensorShape(forward_event_shape)

    def inverse_event_shape(self, event_shape) -> tfp.tf2jax.TensorShape:
      """Returns the shape of the output of `inverse` as a `TensorShape`."""
      self._check_shape("Inverse", event_shape, base_bijector.event_ndims_out)
      inverse_event_shape = jax.eval_shape(
          base_bijector.inverse, jnp.zeros(event_shape)).shape
      return tfp.tf2jax.TensorShape(inverse_event_shape)

    def forward_event_shape_tensor(self, event_shape) -> Array:
      """Returns the shape of the output of `forward` as a `jnp.array`."""
      self._check_shape("Forward", event_shape, base_bijector.event_ndims_in)
      forward_event_shape = jax.eval_shape(
          base_bijector.forward, jnp.zeros(event_shape)).shape
      return jnp.array(forward_event_shape, dtype=jnp.int32)

    def inverse_event_shape_tensor(self, event_shape) -> Array:
      """Returns the shape of the output of `inverse` as a `jnp.array`."""
      self._check_shape("Inverse", event_shape, base_bijector.event_ndims_out)
      inverse_event_shape = jax.eval_shape(
          base_bijector.inverse, jnp.zeros(event_shape)).shape
      return jnp.array(inverse_event_shape, dtype=jnp.int32)

    def forward_log_det_jacobian(
        self, x: Array, event_ndims: Optional[int] = None) -> Array:
      """See `Bijector.forward_log_det_jacobian`."""
      extra_event_ndims = self._check_ndims(
          "Forward", event_ndims, base_bijector.event_ndims_in)
      fldj = base_bijector.forward_log_det_jacobian(x)
      return math.sum_last(fldj, extra_event_ndims)

    def inverse_log_det_jacobian(
        self, y: Array, event_ndims: Optional[int] = None) -> Array:
      """See `Bijector.inverse_log_det_jacobian`."""
      extra_event_ndims = self._check_ndims(
          "Inverse", event_ndims, base_bijector.event_ndims_out)
      ildj = base_bijector.inverse_log_det_jacobian(y)
      return math.sum_last(ildj, extra_event_ndims)

    def _check_ndims(
        self, direction: str, event_ndims: int, expected_ndims: int) -> int:
      """Checks that `event_ndims` are correct and returns any extra ndims."""
      if event_ndims is not None and event_ndims < expected_ndims:
        raise ValueError(f"{direction} `event_ndims` of {self.name} must be at "
                         f"least {expected_ndims} but was passed {event_ndims} "
                         f"instead.")
      return 0 if event_ndims is None else event_ndims - expected_ndims

    def _check_shape(
        self, direction: str, event_shape: Any, expected_ndims: int):
      """Checks that `event_shape` is correct, raising ValueError otherwise."""
      if len(event_shape) < expected_ndims:
        raise ValueError(f"{direction} `event_shape` of {self.name} must have "
                         f"at least {expected_ndims} dimensions, but was "
                         f"{event_shape} which has only {len(event_shape)} "
                         f"dimensions instead.")

    def experimental_compute_density_correction(
        self,
        x: Array,
        tangent_space: TangentSpace,
        backward_compat: bool = True,
        **kwargs) -> Tuple[Array, TangentSpace]:
      """Density correction for this transform wrt the tangent space, at x.

      See `tfp.bijectors.experimental_compute_density_correction`, and
      Radul and Alexeev, AISTATS 2021, “The Base Measure Problem and its
      Solution”, https://arxiv.org/abs/2010.09647.

      Args:
        x: `float` or `double` `Array`.
        tangent_space: `TangentSpace` or one of its subclasses.  The tangent to
          the support manifold at `x`.
        backward_compat: unused
        **kwargs: Optional keyword arguments forwarded to tangent space methods.

      Returns:
        density_correction: `Array` representing the density correction---in log
          space---under the transformation that this Bijector denotes. Assumes
          the Bijector is dimension-preserving.
        space: `TangentSpace` representing the new tangent to the support
          manifold, at `x`.
      """
      del backward_compat
      # We ignore the `backward_compat` flag and always act as though it's
      # true because Distrax bijectors and distributions need not follow the
      # base measure protocol from TFP. This implies that we expect to return
      # the `FullSpace` tangent space.
      return tangent_space.transform_dimension_preserving(x, self, **kwargs)

  return TFPCompatibleBijector()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `normal.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import normal
from distrax._src.utils import equivalence
import jax
import jax.numpy as jnp
import numpy as np


class NormalTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(normal.Normal)

  @parameterized.named_parameters(
      ('1d std normal', (0, 1)),
      ('2d std normal', (np.zeros(2), np.ones(2))),
      ('rank 2 std normal', (np.zeros((3, 2)), np.ones((3, 2)))),
      ('broadcasted loc', (0, np.ones(3))),
      ('broadcasted scale', (np.ones(3), 1)),
  )
  def test_event_shape(self, distr_params):
    super()._test_event_shape(distr_params, dict())

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d std normal, no shape', (0, 1), ()),
      ('1d std normal, int shape', (0, 1), 1),
      ('1d std normal, 1-tuple shape', (0, 1), (1,)),
      ('1d std normal, 2-tuple shape', (0, 1), (2, 2)),
      ('2d std normal, no shape', (np.zeros(2), np.ones(2)), ()),
      ('2d std normal, int shape', ([0, 0], [1, 1]), 1),
      ('2d std normal, 1-tuple shape', (np.zeros(2), np.ones(2)), (1,)),
      ('2d std normal, 2-tuple shape', ([0, 0], [1, 1]), (2, 2)),
      ('rank 2 std normal, 2-tuple shape', (np.zeros((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted loc', (0, np.ones(3)), (2, 2)),
      ('broadcasted scale', (np.ones(3), 1), ()),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_shape(distr_params, dict(), sample_shape)

  @chex.all_variants
  @jax.numpy_rank_promotion('raise')
  @parameterized.named_parameters(
      ('1d std normal, no shape', (0, 1), ()),
      ('1d std normal, int shape', (0, 1), 1),
      ('1d std normal, 1-tuple shape', (0, 1), (1,)),
      ('1d std normal, 2-tuple shape', (0, 1), (2, 2)),
      ('2d std normal, no shape', (np.zeros(2), np.ones(2)), ()),
      ('2d std normal, int shape', ([0, 0], [1, 1]), 1),
      ('2d std normal, 1-tuple shape', (np.zeros(2), np.ones(2)), (1,)),
      ('2d std normal, 2-tuple shape', ([0, 0], [1, 1]), (2, 2)),
      ('rank 2 std normal, 2-tuple shape', (np.zeros((3, 2)), np.ones(
          (3, 2))), (2, 2)),
      ('broadcasted loc', (0, np.ones(3)), (2, 2)),
      ('broadcasted scale', (np.ones(3), 1), ()),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_sample_and_log_prob(
        dist_args=distr_params,
        dist_kwargs=dict(),
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d dist, 1d value', (0, 1), 1),
      ('1d dist, 2d value', (0., 1.), np.array([1, 2])),
      ('2d dist, 1d value', (np.zeros(2), np.ones(2)), 1),
      ('2d broadcasted dist, 1d value', (np.zeros(2), 1), 1),
      ('2d dist, 2d value', (np.zeros(2), np.ones(2)), np.array([1, 2])),
      ('1d dist, 1d value, edge case', (0, 1), 200),
  )
  def test_method_with_input(self, distr_params, value):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    value = np.asarray(value, dtype=np.float32)
    for method in ['log_prob', 'prob', 'cdf', 'log_cdf', 'survival_function',
                   'log_survival_function']:
      with self.subTest(method):
        super()._test_attribute(
            attribute_string=method,
            dist_args=distr_params,
            dist_kwargs={},
            call_args=(value,),
            assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('entropy', (0., 1.), 'entropy'),
      ('mean', (0, 1), 'mean'),
      ('mean from 1d params', ([-1, 1], [1, 2]), 'mean'),
      ('variance', (0, 1), 'variance'),
      ('variance from np params', (np.ones(2), np.ones(2)), 'variance'),
      ('stddev', (0, 1), 'stddev'),
      ('stddev from rank 2 params', (np.ones((2, 3)), np.ones(
          (2, 3))), 'stddev'),
      ('mode', (0, 1), 'mode'),
  )
  def test_method(self, distr_params, function_string):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    super()._test_attribute(
        function_string,
        distr_params,
        assertion_fn=self.assertion_fn(rtol=1e-2))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('no broadcast', ([0., 1., -0.5], [0.5, 1., 1.5])),
      ('broadcasted loc', (0.5, [0.5, 1., 1.5])),
      ('broadcasted scale', ([0., 1., -0.5], 0.8)),
  )
  def test_median(self, distr_params):
    distr_params = (np.asarray(distr_params[0], dtype=np.float32),
                    np.asarray(distr_params[1], dtype=np.float32))
    dist = self.distrax_cls(*distr_params)
    self.assertion_fn(rtol=1e-2)(self.variant(dist.median)(), dist.mean())

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={
            'loc': np.random.randn(4, 1, 2),
            'scale': np.asarray([[0.8, 0.2], [0.1, 1.2], [1.4, 3.1]]),
        },
        dist2_kwargs={
            'loc': np.random.randn(3, 2),
            'scale': 0.1 + np.random.rand(4, 1, 2),
        },
        assertion_fn=self.assertion_fn(rtol=3e-2))

  def test_jittable(self):
    super()._test_jittable((np.zeros((3,)), np.ones((3,))))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
      ('ellipsis', (Ellipsis, -1)),
  )
  def test_slice(self, slice_):
    loc = jnp.array(np.random.randn(3, 4, 5))
    scale = jnp.array(np.random.randn(3, 4, 5))
    dist = self.distrax_cls(loc=loc, scale=scale)
    self.assertion_fn(rtol=1e-2)(dist[slice_].mean(), loc[slice_])

  def test_slice_different_parameterization(self):
    loc = jnp.array(np.random.randn(4))
    scale = jnp.array(np.random.randn(3, 4))
    dist = self.distrax_cls(loc=loc, scale=scale)
    self.assertion_fn(rtol=1e-2)(dist[0].mean(), loc)  # Not slicing loc.
    self.assertion_fn(rtol=1e-2)(dist[0].stddev(), scale[0])

  def test_vmap_inputs(self):
    def log_prob_sum(dist, x):
      return dist.log_prob(x).sum()

    dist = normal.Normal(
        jnp.arange(3 * 4 * 5).reshape((3, 4, 5)), jnp.ones((3, 4, 5)))
    x = jnp.zeros((3, 4, 5))

    with self.subTest('no vmap'):
      actual = log_prob_sum(dist, x)
      expected = dist.log_prob(x).sum()
      self.assertion_fn()(actual, expected)

    with self.subTest('axis=0'):
      actual = jax.vmap(log_prob_sum, in_axes=0)(dist, x)
      expected = dist.log_prob(x).sum(axis=(1, 2))
      self.assertion_fn()(actual, expected)

    with self.subTest('axis=1'):
      actual = jax.vmap(log_prob_sum, in_axes=1)(dist, x)
      expected = dist.log_prob(x).sum(axis=(0, 2))
      self.assertion_fn()(actual, expected)

  def test_vmap_outputs(self):
    def summed_dist(loc, scale):
      return normal.Normal(loc.sum(keepdims=True), scale.sum(keepdims=True))

    loc = jnp.arange((3 * 4 * 5)).reshape((3, 4, 5))
    scale = jnp.ones((3, 4, 5))

    actual = jax.vmap(summed_dist)(loc, scale)
    expected = normal.Normal(
        loc.sum(axis=(1, 2), keepdims=True),
        scale.sum(axis=(1, 2), keepdims=True))

    np.testing.assert_equal(actual.batch_shape, expected.batch_shape)
    np.testing.assert_equal(actual.event_shape, expected.event_shape)

    x = jnp.array([[[1]], [[2]], [[3]]])
    self.assertion_fn(rtol=1e-6)(actual.log_prob(x), expected.log_prob(x))


if __name__ == '__main__':
  absltest.main()
